import torch
from torchvision.models import _utils
import torch.nn as nn
import types

'''-----------------模型组合-----------------------'''


def x_model_group():
    model = nn.Sequential()
    model.add_module('conv', nn.Conv2d(3, 3, 3))
    model.add_module('batchnorm', nn.BatchNorm2d(3))
    model.add_module('activation_layer', nn.ReLU())

    model = nn.Sequential(
        nn.Conv2d(3, 3, 3),
        nn.BatchNorm2d(3),
        nn.ReLU()
    )

    from collections import OrderedDict
    model = nn.Sequential(OrderedDict([
        ('conv', nn.Conv2d(3, 3, 3)),
        ('batchnorm', nn.BatchNorm2d(3)),
        ('activation_layer', nn.ReLU())
    ]))

    # ModuleList 类似 list ，内部没有实现 forward 函数
    model = nn.ModuleList([nn.Linear(3, 4),
                           nn.ReLU(),
                           nn.Linear(4, 2)])


class FNoProcessing(nn.Module):
    '''
    不处理的模型
    '''

    def __init__(self):
        super().__init__()

    def forward(self, inputs):
        return inputs


'''-----------------模型提取双-----------------------'''


class ModelOutsUtils(nn.Module):

    def __init__(self, submodule, ret_name_dict) -> None:
        '''
        只支持 层名为顶层
        '''
        super().__init__()
        self.layer_out = _utils.IntermediateLayerGetter(submodule, ret_name_dict)

    def forward(self, inputs):
        out = self.layer_out(inputs)
        out = list(out.outs())
        # torch.Size([1, 1024, 26, 26])
        return out


class ModelOuts4Resnet(nn.Module):

    def __init__(self, model, dims_out=(512, 1024, 2048)):
        '''
        r18 dims_out = (128, 256, 512)
        r50 dims_out = (512, 1024, 2048)
        :param model:
        :param dims_out:
        '''
        super().__init__()
        self.dims_out = dims_out

        self.model_hook = model
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        # self.model_hook.classifier = NoneLayer()  # 直接输出层
        self.model_hook.layer3[0].conv1.register_forward_hook(self.fun_layer1)
        self.model_hook.layer4[0].conv1.register_forward_hook(self.fun_layer2)
        # model.features[18][2].register_forward_hook(self.fun_layer3)

        self.dims_out = dims_out
        self.out_layout1, self.out_layout2 = [0] * 2

    def foverwrite(self, model, x):
        # 重写流程 model 替换self
        x = model.conv1(x)
        x = model.bn1(x)
        x = model.relu(x)
        x = model.maxpool(x)

        x = model.layer1(x)
        x = model.layer2(x)
        x = model.layer3(x)
        x = model.layer4(x)
        return x

    def fun_layer1(self, module, input, output):
        '''input是默认tuple '''
        self.out_layout1 = input[0]

    def fun_layer2(self, module, input, output):
        self.out_layout2 = input[0]

    def forward(self, inputs):
        outs = self.model_hook(inputs)
        # torch.Size([10, 192, 52, 52]))  torch.Size([10, 576, 26, 26]) torch.Size([10, 1280, 13, 13])
        return self.out_layout1, self.out_layout2, outs
        # return hook

    def freeze_stage(self, stage):
        ''' 冻结resnet所有 '''
        assert stage >= 0 and stage <= 4
        self.model_hook.bn1.eval()
        for m in [self.model_hook.conv1, self.model_hook.bn1]:
            for param in m.parameters():
                param.requires_grad = False

        for i in range(1, stage + 1):
            layer = getattr(self.model_hook, 'layer{}'.format(i))
            layer.eval()
            for param in layer.parameters():
                param.requires_grad = False


class ModelOuts4Resnet_4(nn.Module):

    def __init__(self, model, dims_out=(256, 512, 1024, 2048)):
        super().__init__()
        assert len(dims_out) == 4
        self.model_hook = model
        # 1. 修改(原始类的)方法
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        # self.model_hook.classifier = NoneLayer()  # 直接输出层
        # 2. 在指定层注册侦听方法
        self.model_hook.layer2[0].conv1.register_forward_hook(self.fun_layer1)
        self.model_hook.layer3[0].conv1.register_forward_hook(self.fun_layer2)
        self.model_hook.layer4[0].conv1.register_forward_hook(self.fun_layer3)
        # model.features[18][2].register_forward_hook(self.fun_layer3)

        self.dims_out = dims_out
        self.out_layout1, self.out_layout2 = [0] * 2

    def foverwrite(self, model, x):
        # 重写流程 model 替换self
        x = model.conv1(x)
        x = model.bn1(x)
        x = model.relu(x)
        x = model.maxpool(x)

        x = model.layer1(x)
        x = model.layer2(x)
        x = model.layer3(x)
        x = model.layer4(x)
        return x

    def fun_layer1(self, module, input, output):
        '''input是默认tuple '''
        self.out_layout1 = input[0]

    def fun_layer2(self, module, input, output):
        self.out_layout2 = input[0]

    def fun_layer3(self, module, input, output):
        self.out_layout3 = input[0]

    def forward(self, inputs):
        outs = self.model_hook(inputs)
        # torch.Size([10, 192, 52, 52]))  torch.Size([10, 576, 26, 26]) torch.Size([10, 1280, 13, 13])
        return self.out_layout1, self.out_layout2, self.out_layout3, outs
        # return hook


class ModelOuts4Densenet121(nn.Module):

    def __init__(self, backbone, layer_name, ret_name_dict, dims_out=[512, 1024, 1024]):
        super().__init__()
        submodule = backbone._modules[layer_name]
        self.outs_model = ModelOutsUtils(submodule, ret_name_dict)
        self.dims_out = dims_out

    def forward(self, inputs):
        out1, out2, out3 = self.outs_model(inputs)
        return out1, out2, out3


class ModelOuts4Mobilenet_v2(nn.Module):
    def __init__(self, model):
        super().__init__()
        del model.classifier
        self.model_hook = model
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        # self.model_hook.classifier = NoneLayer()  # 直接输出层
        self.model_hook.features[7].conv[1][0].register_forward_hook(self.fun_layer1)
        self.model_hook.features[14].conv[1][0].register_forward_hook(self.fun_layer2)
        # model.features[18][2].register_forward_hook(self.fun_layer3)

        self.dims_out = [192, 576, 1280]
        self.out_layout1, self.out_layout2 = [0] * 2

    def foverwrite(self, model, x):
        # 重写流程
        x = model.features(x)
        return x

    def fun_layer1(self, module, input, output):
        self.out_layout1 = input[0]

    def fun_layer2(self, module, input, output):
        self.out_layout2 = input[0]

    def forward(self, inputs):
        outs = self.model_hook(inputs)
        # torch.Size([10, 192, 52, 52]))  torch.Size([10, 576, 26, 26]) torch.Size([10, 1280, 13, 13])
        return self.out_layout1, self.out_layout2, outs
        # return hook


class ModelOuts4DarkNet19(nn.Module):
    def __init__(self, model, dims_out=(256, 512, 1024)):
        super().__init__()
        del model.conv_7
        del model.avgpool
        self.model_hook = model
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        self.dims_out = dims_out

    def foverwrite(self, model, x):
        x = model.conv_1(x)
        x = model.conv_2(x)
        x = model.conv_3(x)
        ceng1 = model.conv_4(x)
        ceng2 = model.conv_5(model.maxpool_4(ceng1))
        ceng3 = model.conv_6(model.maxpool_5(ceng2))
        return ceng1, ceng2, ceng3

    def forward(self, inputs):
        outs = self.model_hook(inputs)
        return outs


class ModelOuts4DarkNet19_Tiny(nn.Module):
    def __init__(self, model, dims_out=(128, 256, 512)):
        super().__init__()
        del model.conv_6
        del model.avgpool
        self.model_hook = model
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        self.dims_out = dims_out

    def foverwrite(self, model, x):
        x = model.conv_1(x)
        x = model.conv_2(x)
        ceng1 = model.conv_3(x)
        ceng2 = model.conv_4(ceng1)
        ceng3 = model.conv_5(ceng2)
        return ceng1, ceng2, ceng3

    def forward(self, inputs):
        outs = self.model_hook(inputs)
        return outs


class ModelOuts4DarkNet53(nn.Module):
    def __init__(self, model, dims_out=(256, 512, 1024)):
        super().__init__()
        del model.avgpool
        del model.fc
        self.model_hook = model
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        self.dims_out = dims_out

    def foverwrite(self, model, x):
        x = model.layer_1(x)
        x = model.layer_2(x)
        ceng1 = model.layer_3(x)
        ceng2 = model.layer_4(ceng1)
        ceng3 = model.layer_5(ceng2)
        return ceng1, ceng2, ceng3

    def forward(self, inputs):
        outs = self.model_hook(inputs)
        return outs


class ModelOuts4Shufflenet(nn.Module):

    def __init__(self, model, dims_out=(116, 232, 464)):
        '''
        :param model:
        :param dims_out:
            if model_size == '0.5x':
                self._stage_out_channels = [24, 48, 96, 192, 1024]
            elif model_size == '1.0x':
                self._stage_out_channels = [24, 116, 232, 464, 1024]
            elif model_size == '1.5x':
                self._stage_out_channels = [24, 176, 352, 704, 1024]
            elif model_size == '2.0x':
                self._stage_out_channels = [24, 244, 488, 976, 2048]
        '''
        super().__init__()
        self.dims_out = dims_out

        self.model_hook = model
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        # 注册最后的两层 取input
        self.model_hook.stage3[0].branch1[0].register_forward_hook(self.fun_layer1)
        self.model_hook.stage4[0].branch1[0].register_forward_hook(self.fun_layer2)

        self.dims_out = dims_out
        self.out_layout1, self.out_layout2 = [0] * 2  # 赋初值

    def foverwrite(self, model, x):
        # 重写流程 model 替换self
        x = model.conv1(x)
        x = model.maxpool(x)
        x = model.stage2(x)  # 输出 torch.Size([2, 116, 52, 52])
        x = model.stage3(x)
        x = model.stage4(x)
        return x

    def fun_layer1(self, module, input, output):
        '''input是默认tuple '''
        self.out_layout1 = input[0]

    def fun_layer2(self, module, input, output):
        self.out_layout2 = input[0]

    def forward(self, inputs):
        outs = self.model_hook(inputs)
        # torch.Size([10, 192, 52, 52]))  torch.Size([10, 576, 26, 26]) torch.Size([10, 1280, 13, 13])
        return self.out_layout1, self.out_layout2, outs
        # return hook

    def freeze_bn(self):
        ''' 冻结模型的BN层 '''
        for layer in self.model_hook.modules():
            if isinstance(layer, nn.BatchNorm2d):
                layer.eval()

    def freeze_stage(self, stage):
        ''' 冻结resnet所有 '''
        assert stage >= 0 and stage <= 4
        self.model_hook.bn1.eval()
        for m in [self.model_hook.conv1, self.model_hook.bn1]:
            for param in m.parameters():
                param.requires_grad = False

        for i in range(1, stage + 1):
            layer = getattr(self.model_hook, 'layer{}'.format(i))
            layer.eval()
            for param in layer.parameters():
                param.requires_grad = False


'''-----------------单输出---------------------'''


class ModelOut4DarkNet19(nn.Module):
    def __init__(self, model):
        super().__init__()
        del model.conv_7
        del model.avgpool
        self.model = model
        # 重写流程
        self.model._forward_impl = types.MethodType(self.foverwrite, model)
        self.dim_out = 1024

    def foverwrite(self, model, x):
        x = model.conv_1(x)
        x = model.conv_2(x)
        x = model.conv_3(x)
        x = model.conv_4(x)
        x = model.conv_5(model.maxpool_4(x))
        x = model.conv_6(model.maxpool_5(x))

        # x = self.avgpool(x)
        # x = self.conv_7(x)
        # x = x.view(x.size(0), -1)
        return x

    def forward(self, inputs):
        ous = self.model(inputs)
        return ous


class ModelOut4Resnet18(nn.Module):
    def __init__(self, model):
        super().__init__()
        del model.avgpool
        del model.fc
        self.model = model
        # 重写流程
        self.model._forward_impl = types.MethodType(self.foverwrite, model)
        self.dim_out = 512

    def foverwrite(self, model, x):
        # 重写流程
        x = model.conv1(x)
        x = model.bn1(x)
        x = model.relu(x)
        x = model.maxpool(x)

        x = model.layer1(x)
        x = model.layer2(x)
        x = model.layer3(x)
        x = model.layer4(x)

        # x = self.avgpool(x)
        # x = torch.flatten(x, 1)
        # x = self.fc(x)
        return x

    def forward(self, inputs):
        ous = self.model(inputs)
        return ous


class ModelOut4Resnet50(ModelOut4Resnet18):

    def __init__(self, model):
        super().__init__(model)
        self.dim_out = 2048


class ModelOut4Mobilenet_v2(nn.Module):
    def __init__(self, model):
        super().__init__()
        del model.classifier
        self.model_hook = model
        # 直接全部改了
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        self.dim_out = 1280
        # model.features[18][2].register_forward_hook(self.fun_layer1)
        # self.out_layout1 = None

    def foverwrite(self, model, x):
        # 重写流程
        x = model.features(x)
        return x

    def forward(self, inputs):
        ous = self.model_hook(inputs)
        return ous


class ModelOut4Mobilenet_v3(nn.Module):
    def __init__(self, model):
        super().__init__()
        del model.linear3
        del model.bn3
        del model.hs3
        del model.linear4
        self.model_hook = model
        # 类中调用 第一参数是self  必须定义一个参数 是第二个参数, 调用传入的是第三个
        self.model_hook._forward_impl = types.MethodType(self.foverwrite, model)
        self.dim_out = 576

    def foverwrite(self, model, x):
        '''

        :param model: MethodType 必须传一个没用的参数
        :param x:
        :return:
        '''
        out = model.hs1(model.bn1(model.conv1(x)))
        out = model.bneck(out)
        out = model.hs2(model.bn2(model.conv2(out)))
        # out = F.avg_pool2d(out, 7)
        # out = out.view(out.size(0), -1)
        # out = self.hs3(self.bn3(self.linear3(out)))
        # out = self.linear4(out)
        return out

    def forward(self, inputs):
        ous = self.model_hook(inputs)
        return ous


if __name__ == '__main__':
    from torchvision import models

    '''通过 层层遍历 提取'''
    # model = models.densenet121(pretrained=True)
    # # f_look(model, input=(1, 3, 416, 416))
    # # conv 可以取 in_channels 不支持数组层
    # dims_out = [512, 1024, 1024]
    # print(dims_out)
    # ret_name_dict = {'denseblock2': 1, 'denseblock3': 2, 'denseblock4': 3}
    # model = ModelOut4Densenet121(model, 'features', ret_name_dict)
    # f_look_model(model, input=(1, 3, 416, 416))

    # 这个分不出来
    # model = models.mobilenet_v2(pretrained=True)
    # f_look(model, input=(1, 3, 416, 416))
    # model = ModelOut4Mobilenet_v2(model)
    # dims_out = [192, 576, 1280]
    # f_look_model(model, input=(1, 3, 416, 416))

    # model = models.resnet50(pretrained=True)
    # model = models.resnext50_32x4d(pretrained=True)
    # dims_out = (512, 1024, 2048)
    model = models.resnet18(pretrained=True)
    dims_out = (128, 256, 512)
    model = ModelOuts4Resnet(model, dims_out)
    # f_look_model(model, input=(1, 3, 416, 416))

    x = torch.zeros((1, 3, 416, 416))
    outs = model(x)
    print(outs)

    # model = models.resnext50_32x4d(pretrained=True)
    # model = models.densenet161(pretrained=True)
    # ret_name_dict = {'layer2': 1, 'layer3': 2, 'layer4': 3}
    # model = Output4Return(model, return_layers)

    # dims_out = [512, 1024, 2048]
    # f_look(model, input=(1, 3, 416, 416))

    # model = models.resnet50(pretrained=True)
    # model = models.wide_resnet50_2(pretrained=True)
    # model = models.mnasnet0_5(pretrained=True)
    # model = models.mobilenet_v2(pretrained=True)
    # model = ModelOuts4Mobilenet_v2(model)
    # save_weight('.', model, '123')
    # load_weight('./123-1_None.pth', model)

    # checkpoint = torch.load('./123-1_None.pth')
    # pretrained_dict = checkpoint['model']
    # keys_missing, keys_unexpected = model.load_state_dict(pretrained_dict, strict=False)
    # if len(keys_missing) > 0 or len(keys_unexpected):
    #     flog.error('missing_keys %s', keys_missing)
    #     flog.error('unexpected_keys %s', keys_unexpected)
    # else:
    #     flog.info('完成')

    # dims_out = [240, 576, 1280]
    # dims_out = [144, 288, 1280]
    # layer1 = model.layers[10][0].layers[3]
    # layer2 = model.layers[12][0].layers[3]
    # layer3 = model.layers[10][0].layers[3]
    # ModelOuts4Mnasnet(model)
    #
    # hook1_obj = model.layers[10][0].layers[3].register_forward_hook(fun1)  # 提取输入
    # hook2_obj = model.layers[12][0].layers[3].register_forward_hook(fun1)  # 提取输入
    # hook3_obj = model.layers[16].register_forward_hook(fun1)  # 提取输入
    # hook_obj.remove()
    # del model.classifier
    # ModelOut4Mnasnet1_0(model,)
    # print(hook_obj)
    # f_look_model(model, input=(10, 3, 416, 416))
    pass



import torch
import torch.nn.functional as F
import numpy as np


def fquality_focal_loss(pcls, gcls, score, beta=2.0):
    '''
    支持0~1连续标签
    :param pcls: torch.Size([3, 2100, 80])
    :param gcls: torch.Size([3, 2100, 80])
    :param score: torch.Size([3, 2100])
    :param beta:
    :return: torch.Size([3, 2100, 80])
    '''
    gcls_score = gcls * score.unsqueeze(-1)
    loss_bce = F.binary_cross_entropy_with_logits(pcls, gcls_score, reduction='none')
    scale_factor = (pcls.sigmoid() - gcls_score).abs().pow(beta)
    loss = loss_bce * scale_factor
    return loss


def focalloss_fast(pcls_sigmoid, gcls, alpha=0.25, gamma=2.0):
    '''
    gcls只支持0或1标签
    '''
    eps = torch.finfo(torch.float16).eps
    pcls_sigmoid = pcls_sigmoid.clamp(min=eps, max=1 - eps)
    pt = pcls_sigmoid * gcls + (1.0 - pcls_sigmoid) * (1.0 - gcls)
    w = alpha * gcls + (1.0 - alpha) * (1.0 - gcls)
    loss = -w * torch.pow((1.0 - pt), gamma) * pt.log()
    return loss


def focalloss_center(pcls_sigmoid, gcls, alpha=2., beta=4.):
    '''
    支持0~1的连续标签 好像有损失上限
    '''
    eps = torch.finfo(torch.float16).eps
    pcls_sigmoid = pcls_sigmoid.clamp(min=eps, max=1 - eps)
    mask_pos_3d = gcls == 1
    mask_neg_3d = gcls != 1
    l_pos = -mask_pos_3d.float() * ((1.0 - pcls_sigmoid) ** alpha) * torch.log(pcls_sigmoid)
    l_neg = -mask_neg_3d.float() * ((1 - gcls) ** beta) * (pcls_sigmoid ** alpha) * torch.log(1.0 - pcls_sigmoid)
    return l_pos + l_neg


def fdistribution_focal_loss(pred, label, reduction='none'):
    '''
    通过 ce 使分布靠近GT的左右值   通过不同的权重使其回归到指定位置
    :param pred:  torch.Size([27, 8, 4])
    :param label: torch.Size([27, 4])
    :return:
        torch.Size([27, 4])

    pred:  2d:[n,8]  3d:[b,8,n]
    label: 2d:[n]    3d:[b,n] 值必须在在[0,7) 左闭右开
    # ---- 2d
    preg = torch.randint(0, 8, (3, 8)) + torch.rand((3, 8))
    print(preg)
    greg = torch.randint(0, 7, (3,)) + torch.rand(3)
    print(greg)
    loss = distribution_focal_loss(preg, greg)
    print(loss)

    # ---- 3d
    preg = torch.randint(0, 8, (3, 2, 8)) + torch.rand((3, 2, 8))
    preg = preg.permute(0, 2, 1) # (3, 2, 8) -> (3, 8, 2)
    print(preg)
    greg = torch.randint(0, 7, (3, 2)) + torch.rand(3, 2)
    print(greg)
    loss = distribution_focal_loss(preg, greg)
    print(loss)
    '''
    disl = label.long()  # 向下取整
    disr = disl + 1  # 向上取整

    # GT靠哪边哪边的权重就大
    wl = disr.float() - label  # GT距离左边的权限
    wr = label - disl.float()

    # torch.Size([3, 8400, 8])  ^^ torch.Size([3, 8400])
    loss = F.cross_entropy(pred, disl, reduction='none') * wl \
           + F.cross_entropy(pred, disr, reduction='none') * wr
    return loss


''' -------------------------  ------------------------------ '''


# def quality_focal_loss(pred, target, beta=2.0):
#     assert len(target) == 2, """target for QFL must be a tuple of two elements,
#         including category label and quality label, respectively"""
#     # label denotes the category id, score denotes the quality score
#     label, score = target
#
#     # negatives are supervised by 0 quality score
#     pred_sigmoid = pred.sigmoid()
#     scale_factor = pred_sigmoid
#     zerolabel = scale_factor.new_zeros(pred.shape)
#     loss = F.binary_cross_entropy_with_logits(
#         pred, zerolabel, reduction='none') * scale_factor.pow(beta)
#
#     # FG cat_id: [0, num_classes -1], BG cat_id: num_classes
#     bg_class_ind = pred.size(1)
#     pos = torch.nonzero((label >= 0) & (label < bg_class_ind), as_tuple=False).squeeze(1)
#     pos_label = label[pos].long()
#     # positives are supervised by bbox quality (IoU) score
#     scale_factor = score[pos] - pred_sigmoid[pos, pos_label]
#     loss[pos, pos_label] = F.binary_cross_entropy_with_logits(
#         pred[pos, pos_label], score[pos],
#         reduction='none') * scale_factor.abs().pow(beta)
#
#     loss = loss.sum(dim=1, keepdim=False)
#     return loss
def t_fl():
    '''
    三种focalloss测试
    '''
    pcls = torch.arange(12, dtype=torch.float).reshape(1, 3, 4)
    gcls = torch.zeros_like(pcls)
    gcls[0][0][1] = 1
    gcls[0][1][2] = 1
    gcls[0][2][0] = 1
    # score = torch.rand_like(gcls)
    score = torch.ones_like(gcls)
    print(pcls.sigmoid())
    print(gcls * score)
    # loss1 = quality_focal_loss(pcls, gcls, score)
    # print(loss1)
    loss_gfl = fquality_focal_loss(pcls, gcls, score)
    print(loss_gfl)
    loss_center = focalloss_center(pcls.sigmoid(), gcls * score)
    print(loss_center)
    loss_fl = focalloss_fast(pcls.sigmoid(), gcls * score)
    print(loss_fl)


if __name__ == '__main__':
    torch.manual_seed(3)
    np.set_printoptions(suppress=True)
    torch.set_printoptions(linewidth=320, sci_mode=False, precision=5, profile='long')


    # t_fl()
    pass



import datetime
import logging
import logging.config
import os
from logging import handlers

import colorlog

from tools.f_general import get_path_root

'''
在项目根目录建立 logs 文件夹
args=(os.path.abspath(os.getcwd() + "/info.log"),"midnight", 1, 6,'utf-8')
每一天午夜12点将当天的日志转存到一份新的日志文件中，并且加上时间戳后缀，最多保存6个文件
'''


def Singleton(cls):
    # 单例
    _instance = {}

    def _singleton(*args, **kargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kargs)
        return _instance[cls]

    return _singleton


def get_logger(name='root'):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # fmt = '%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s'
    # formatter = logging.Formatter(fmt)
    # black, red, green, yellow, blue, purple, cyan and white {color}，fg_ {color}，bg_ {color}：前景色和背景色
    log_colors_config = {
        'DEBUG': 'blue',  # 蓝色
        # 'INFO': 'green',  # 绿色
        'INFO': 'cyan',  # 蓝绿
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'red,bg_white',
    }
    formatter = colorlog.ColoredFormatter(
        '%(log_color)s[%(asctime)s] [%(filename)s:%(lineno)d] [%(module)s:%(funcName)s] [%(levelname)s]- %(message)s',
        log_colors=log_colors_config)  # 日志输出格式

    sh = logging.StreamHandler()
    sh.setFormatter(formatter)
    # sh.setLevel(logging.DEBUG)
    logger.addHandler(sh)

    '''
    实例化TimedRotatingFileHandler
        #interval是时间间隔，backupCount是备份文件的个数，如果超过这个个数，就会自动删除，when是间隔的时间单位，单位有以下几种：
        # S 秒
        # M 分
        # H 小时、
        # D 天、
        # W 每星期（interval==0时代表星期一）
        # midnight 每天凌晨
    '''
    path_root = os.path.join(get_path_root(), 'logs')
    if not os.path.exists(path_root):
        os.makedirs(path_root)
    th = handlers.TimedRotatingFileHandler(
        filename=os.path.join(path_root, datetime.datetime.now().strftime('%Y_%m_%d') + '.log'),
        when='D',
        backupCount=5, encoding='utf-8')
    th.setFormatter(formatter)
    logger.addHandler(th)

    # fh = RotatingFileHandler(
    #     filename=os.path.join(get_path_root(), 'logs', datetime.datetime.now().strftime('%Y_%m_%d') + '.log'),
    #     mode='a', maxBytes=1024 * 1024 * 5, backupCount=5,
    #     encoding='utf-8')  # 使用RotatingFileHandler类，滚动备份日志
    # fh.setLevel(logging.CRITICAL)
    # fh.setFormatter(formatter)
    # print(logger.handlers)
    # logger.removeHandler(ch)
    return logger


flog = get_logger(__name__)  # 返回一个叫__name__ 的obj，并应用默认的日志级别、Handler和Formatter设置

# flog=Log()
# flog1=Log()
# print(id(flog))
# print(id(flog1))

if __name__ == '__main__':
    # flog.debug('一个连接只需一个 %s', get_path_root)
    flog.debug('多个连接无需   %s%s', [1, 2, {123}], get_path_root())
    flog.info(123)
    flog.warning('多个连接无需   %s%s', [1, 2, {123}], get_path_root())
    flog.error(123)
    flog.critical(123)
    pass



import os
import time

import torch
import torch.nn as nn
from torch.cuda.amp import GradScaler, autocast

from tools.GLOBAL_LOG import flog


class FModelBase(nn.Module):
    def __init__(self, cfg, net, losser, preder):
        super(FModelBase, self).__init__()
        self.cfg = cfg
        self.net = net
        self.losser = losser
        self.preder = preder

    def forward(self, reses):
        imgs_ts_4d, targets = reses
        outs = self.net(imgs_ts_4d)

        ''' 多尺度图片归一化支持 '''
        device = imgs_ts_4d.device
        batch, c, h, w = imgs_ts_4d.shape
        toones_wh_ts_input = []
        if targets is not None and 'toone' in targets[0]:  # 多尺度归一
            for target in targets:
                toones_wh_ts_input.append(torch.tensor(target['toone'], device=device))
            toones_wh_ts_input = torch.stack(toones_wh_ts_input, 0)
        else:
            toones_wh_ts_input = torch.empty(batch, 2, device=device, dtype=torch.float)  # 实际为准
            toones_wh_ts_input[:, :] = torch.tensor(imgs_ts_4d.shape[2:4][::-1], device=device)

        if self.training:
            if targets is None:
                raise ValueError("In training mode, targets should be passed")
            loss_total, log_dict = self.losser(outs, targets, imgs_ts_4d, toones_wh_ts_input)
            '''------验证loss 待扩展------'''

            return loss_total, log_dict
        else:
            with torch.no_grad():  # 这个没用
                # outs模型输出  input_image
                ids_batch, p_boxes_ltrb, p_keypoints, p_labels, p_scores = self.preder(outs, imgs_ts_4d, targets,
                                                                                       toones_wh_ts_input)
            return ids_batch, p_boxes_ltrb, p_keypoints, p_labels, p_scores


class AverageMeter(object):
    """ Computes ans stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0.
        self.avg = 0.
        self.sum = 0.
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class FitExecutor():
    def __init__(self, cfg, model, fun_loss, save_weight_name, path_save_weight,
                 optimizer=None, dataloader_train=None, lr_scheduler=None,
                 end_epoch=None, is_mixture_fit=True, lr_val_base=1e-3,
                 device=torch.device('cpu'), is_writer=False,
                 num_save_interval=10, print_freq=10,
                 dataloader_val=None, val_freq=1,
                 ) -> None:
        '''

        :param model:
        :param fun_loss: 回调 loss 处理
        :param save_weight_name:
        :param path_save_weight:
        :param optimizer:
        :param dataloader_train:
        :param lr_scheduler:
        :param end_epoch:
        :param is_mixture_fit:
        :param lr_val_base:
        :param device:
        :param is_writer:
        :param num_save_interval:
        :param print_freq:
        :param dataloader_val:
        :param val_freq: 验证频率
        '''
        super().__init__()
        self.dataloader_train = dataloader_train
        self.model = model
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.end_epoch = end_epoch
        self.is_mixture_fit = is_mixture_fit
        self.lr_val_base = lr_val_base
        self.device = device

        self.fun_loss = fun_loss
        self.cfg = cfg

        # 验证
        self.dataloader_val = dataloader_val
        self.val_freq = val_freq

        self.save_weight_name = save_weight_name
        self.path_save_weight = path_save_weight
        self.num_save_interval = num_save_interval
        self.print_freq = print_freq

        if is_writer:
            flog.debug('---- use tensorboard ---')
            from torch.utils.tensorboard import SummaryWriter
            c_time = time.strftime('%Y-%m-%d_%H_%M_%S', time.localtime(time.time()))
            _path = os.path.join(path_save_weight, c_time)
            os.makedirs(_path, exist_ok=True)
            self.tb_writer = SummaryWriter(_path)
        else:
            self.tb_writer = None

    def frun(self, start_epoch):
        t1 = time.time()
        for epoch in range(start_epoch, self.end_epoch + 1, 1):  # 从1开始
            save_val = None
            ''' ------------------- 训练代码  --------------------- '''
            if self.dataloader_train is not None and self.cfg.IS_TRAIN:
                loss_train_obj = self.ftrain(epoch, t1)
                save_val = loss_train_obj.avg

            ''' ------------------- 验证代码  --------------------- '''
            if self.dataloader_val is not None and self.cfg.IS_VAL and epoch % self.val_freq == 0:
                loss_val_obj = self.fval(epoch)

            if self.dataloader_train is not None and (epoch % self.num_save_interval) == 0:
                print('训练完成正在保存模型...')
                save_weight(
                    path_save=self.path_save_weight,
                    model=self.model,
                    name=self.save_weight_name,
                    loss=save_val,  # 这个正常是一样的有的
                    optimizer=self.optimizer,
                    lr_scheduler=self.lr_scheduler,
                    epoch=epoch)

    def fval(self, epoch):
        print('\n-------------------- 验证开始 开始 %s -------------------------' % epoch)
        losses_val_obj = AverageMeter()
        epoch_size = len(self.dataloader_val)

        self.model.eval()
        for i, reses in enumerate(self.dataloader_val):
            t0 = time.time()
            with torch.no_grad():
                loss_total, log_dict = self.fun_loss(reses)

                if i % (self.print_freq * self.cfg.BATCH_VAL) == 0:
                    show_loss_str = []
                    for k, v, in log_dict.items():
                        show_loss_str.append(
                            "{}: {:.4f} ||".format(k, v)
                        )

                    print('[Val %d/%d][Iter %d/%d/%d][lr %.6f]'
                          '[Loss: all %.2f || %s  time: %.2f]'
                          % (epoch, self.end_epoch, i + 1,
                             epoch_size, (epoch - 1) * epoch_size + i + 1,
                             self.optimizer.param_groups[0]['lr'],
                             losses_val_obj.avg,
                             str(show_loss_str),
                             time.time() - t0,
                             ),
                          flush=True
                          )

            if self.tb_writer is not None:
                iter = epoch_size * (epoch - 1) + i + 1
                # 主进程写入
                for k, v, in log_dict.items():
                    self.tb_writer.add_scalar('loss_iter/val-%s' % k, v, iter)

        return losses_val_obj

    def ftrain(self, epoch, t1):
        # 这里是 epoch
        # show_tb_writer = []
        loss_train_obj = AverageMeter()
        print('-------------------- fit_train 开始 %s -------------------------' % epoch)
        scaler = GradScaler(enabled=self.is_mixture_fit)
        epoch_size = len(self.dataloader_train)

        print('\n-------------------- 训练 开始 %s -------------------------' % epoch)
        for i, reses in enumerate(self.dataloader_train):
            t0 = time.time()
            if epoch < 2:
                now_lr = self.lr_val_base * pow((i + epoch * epoch_size) * 1. / (1 * epoch_size), 4)
                for param_group in self.optimizer.param_groups:
                    param_group['lr'] = now_lr
            elif epoch == 1:
                now_lr = self.lr_val_base
                for param_group in self.optimizer.param_groups:
                    param_group['lr'] = now_lr

            with autocast(enabled=self.is_mixture_fit):
                if self.fun_loss is not None:
                    # 这里是回调
                    loss_total, log_dict = self.fun_loss(reses)
                else:
                    loss_total, log_dict = self.model(reses)

            loss_train_obj.update(loss_total.item())

            scaler.scale(loss_total).backward()
            scaler.step(self.optimizer)
            scaler.update()
            self.optimizer.zero_grad()

            if i % (self.print_freq * self.cfg.BATCH_TRAIN) == 0:
                show_loss_str = []
                for k, v, in log_dict.items():
                    show_loss_str.append(
                        "{}: {:.4f} ||".format(k, v)
                    )

                print('[Train %d/%d][Iter %d/%d/%d][lr %.6f]'
                      '[Loss: all %.2f || %s  time: %.2f/%.2f]'
                      % (epoch, self.end_epoch, i + 1,
                         epoch_size, (epoch - 1) * epoch_size + i + 1,
                         self.optimizer.param_groups[0]['lr'],
                         loss_train_obj.avg,
                         str(show_loss_str),
                         t0 - t1,
                         time.time() - t1,
                         ),
                      flush=True
                      )

            if self.tb_writer is not None:
                # 主进程写入   不验证时写
                iter = epoch_size * (epoch - 1) + i + 1
                for k, v, in log_dict.items():
                    self.tb_writer.add_scalar('loss_iter/%s' % k, v, iter)
                self.tb_writer.add_scalar('loss_iter/lr', self.optimizer.param_groups[0]['lr'], iter)

            # 更新时间用于获取 data 时间
            t1 = time.time()

        return loss_train_obj


def save_weight(path_save, model, name, loss=None, optimizer=None, lr_scheduler=None, epoch=0, maps_val=None):
    if path_save and os.path.exists(path_save):
        sava_dict = {
            'model': model.state_dict(),
            'optimizer': optimizer.state_dict() if optimizer else None,
            'lr_scheduler': lr_scheduler.state_dict() if lr_scheduler else None,
            'epoch': epoch}
        if maps_val is not None:
            if loss is not None:
                l = round(loss, 2)
            else:
                l = ''
            file_weight = os.path.join(path_save, (name + '-{}_{}_{}_{}.pth')
                                       .format(epoch + 1,
                                               l,
                                               'p' + str(round(maps_val[0] * 100, 1)),
                                               'r' + str(round(maps_val[1] * 100, 1)),
                                               ))
        else:
            file_weight = os.path.join(path_save, (name + '-{}_{}.pth').format(epoch + 1, round(loss, 3)))
        torch.save(sava_dict, file_weight)
        flog.info('保存成功 %s', file_weight)


def load_weight(file_weight, model, optimizer=None, lr_scheduler=None,
                device=torch.device('cpu'), is_mgpu=False, ffun=None):
    start_epoch = 1
    if file_weight and os.path.exists(file_weight):
        checkpoint = torch.load(file_weight, map_location=device)

        '''对多gpu的k进行修复'''
        # if 'model' in checkpoint:
        #     pretrained_dict_y = checkpoint['model']
        # else:
        #     pretrained_dict_y = checkpoint

        ''' debug '''
        # if ffun is not None:
        #     pretrained_dict = ffun(pretrained_dict_y)
        # else:
        #     pretrained_dict = pretrained_dict_y

        dd = {}

        # # 多GPU处理
        # ss = 'module.'
        # for k, v in pretrained_dict.items():
        #     if is_mgpu:
        #         if ss not in k:
        #             dd[ss + k] = v
        #         else:
        #             dd = pretrained_dict_y
        #             break
        #             # dd[k] = v
        #     else:
        #         dd[k.replace(ss, '')] = v

        '''重组权重'''
        # load_weights_dict = {k: v for k, v in weights_dict.items()
        #                      if model.state_dict()[k].numel() == v.numel()}

        keys_missing, keys_unexpected = model.load_state_dict(dd, strict=False)
        if len(keys_missing) > 0 or len(keys_unexpected):
            flog.error('missing_keys %s', keys_missing)  # 这个是 model 的属性
            flog.error('unexpected_keys %s', keys_unexpected)  # 这个是 pth 的属性
        if optimizer:
            optimizer.load_state_dict(checkpoint['optimizer'])
        if lr_scheduler and checkpoint['lr_scheduler']:
            lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])

        if 'epoch' in checkpoint:
            start_epoch = checkpoint['epoch']
        else:
            start_epoch = 1

        flog.warning('已加载 feadre 权重文件为 %s', file_weight)
    return start_epoch



import torch.nn as nn
import torch.nn.functional as F


class FPN_out4add(nn.Module):

    def __init__(self, in_channels_list, out_channel=256,
                 use_p5=True, num_outs=5, is_init=False, use_back_conv=True):
        '''

        :param in_channels_list: 写死只支持3层
        :param out_channel:
        :param num_outs: 只支持3或5
        '''
        super(FPN_out4add, self).__init__()
        self.add_module('prj_5', nn.Conv2d(in_channels_list[2], out_channel, kernel_size=1))
        # self.prj_5 = nn.Conv2d(in_channels_list[2], out_channel, kernel_size=1)
        self.prj_4 = nn.Conv2d(in_channels_list[1], out_channel, kernel_size=1)
        self.prj_3 = nn.Conv2d(in_channels_list[0], out_channel, kernel_size=1)

        self.use_back_conv = use_back_conv
        if self.use_back_conv:  # 尺寸不变
            self.conv_5 = nn.Conv2d(out_channel, out_channel, kernel_size=3, padding=1)
            self.conv_4 = nn.Conv2d(out_channel, out_channel, kernel_size=3, padding=1)
            self.conv_3 = nn.Conv2d(out_channel, out_channel, kernel_size=3, padding=1)

        if num_outs == 5:
            # 输出5层才启动
            if use_p5:
                self.conv_out6 = nn.Conv2d(out_channel, out_channel, kernel_size=3, padding=1, stride=2)
            else:
                self.conv_out6 = nn.Conv2d(in_channels_list[2], out_channel, kernel_size=3, padding=1, stride=2)
            self.conv_out7 = nn.Conv2d(out_channel, out_channel, kernel_size=3, padding=1, stride=2)

        if is_init:
            self.init_conv(self)

        self.use_p5 = use_p5
        self.num_outs = num_outs
        # self.apply(finit_conv_kaiming)
        self.out_channel = out_channel

    def init_conv(self, model):
        for m in model.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight)

    def upsamplelike(self, inputs):
        '''

        :param inputs: 第二个是尺寸 size_hw 行列
        :return:
        '''
        src, target = inputs
        #  nn.Upsample 与 F.interpolate 等价
        return F.interpolate(src, size=(target.shape[2], target.shape[3]), mode='nearest')

    def forward(self, inputs):
        C3, C4, C5 = inputs
        P5 = self.prj_5(C5)
        P4 = self.prj_4(C4)
        P3 = self.prj_3(C3)

        P4 = P4 + self.upsamplelike([P5, C4])
        P3 = P3 + self.upsamplelike([P4, C3])

        if self.use_back_conv:
            P3 = self.conv_3(P3)
            P4 = self.conv_4(P4)
            P5 = self.conv_5(P5)

        if self.num_outs == 5:
            P5 = P5 if self.use_p5 else C5
            P6 = self.conv_out6(P5)
            P7 = self.conv_out7(F.relu(P6))
            return [P3, P4, P5, P6, P7]
        else:
            return [P3, P4, P5]


class PAN_out4add(FPN_out4add):

    def __init__(self, in_channels_list, out_channel=256,
                 use_p5=True, num_outs=5, is_init=False, use_back_conv=True):
        '''

        :param in_channels_list: 写死只支持3层
        :param out_channel:
        :param num_outs: 只支持3或5
        '''
        super().__init__(in_channels_list, out_channel, use_p5, num_outs, is_init, use_back_conv)

    def forward(self, inputs):
        C3, C4, C5 = inputs
        P5 = self.prj_5(C5)
        P4 = self.prj_4(C4)
        P3 = self.prj_3(C3)

        P4 = P4 + self.upsamplelike([P5, C4])
        P3 = P3 + self.upsamplelike([P4, C3])

        P4 = P4 + self.upsamplelike([P5, C4])
        P5 = P5 + self.upsamplelike([P4, C5])

        if self.use_back_conv:
            P3 = self.conv_3(P3)
            P4 = self.conv_4(P4)
            P5 = self.conv_5(P5)

        if self.num_outs == 5:
            P5 = P5 if self.use_p5 else C5
            P6 = self.conv_out6(P5)
            P7 = self.conv_out7(F.relu(P6))
            return [P3, P4, P5, P6, P7]
        else:
            return [P3, P4, P5]



import os
import sys

curPath = os.path.abspath(os.path.dirname(__file__))
rootPath = os.path.split(curPath)[0]
sys.path.append(os.path.split(rootPath)[0])

import torch
import torchvision.models as models
from torch import nn
from thop import profile


def f_look_tw(model, input=(1, 3, 416, 416), name='model_look'):
    import tensorwatch as tw
    # 用这个即可---查看网络的统计结果---
    args_pd = tw.model_stats(model, input)
    args_pd.to_excel(name + '.xlsx')
    script_models = torch.jit.trace(model, input)
    script_models.save(name + '_jit' + '.pt')
    print('文件生成成功 %s', name + '.xlsx')


def f_look_summary(model, input=(3, 416, 416), device="cpu"):
    '''
    这个不支持 tuple 输入  没有.size()属性会报错
    :param model:
    :param input:
    :param device:
    :return:
    '''
    from torchsummary import summary
    if not isinstance(input, tuple):
        input = tuple(input)
    summary1 = summary(model, input, device=device)
    print(type(summary1))


def f_calc_flops_params(model, size_hw):
    inputs = torch.randn(1, 3, *size_hw)
    flops, params = profile(model, (inputs,))
    print('flops: ', flops, 'params: ', params)
    return flops, params


if __name__ == '__main__':
    '''
    '''
    # data_inputs_list = [1, 3, 640, 640]
    data_inputs_list = [1, 3, 416, 416]

    torch.random.manual_seed(20201025)  # 3746401707500
    data_inputs_ts = torch.rand(data_inputs_list, dtype=torch.float)

    # model = models.AlexNet()
    # model = models.densenet161(pretrained=True)  # 能力 22.35  6.20  ---top2
    # model = FRebuild4densenet161(model, None)
    # return_layers = {'layer1': 1, 'layer2': 2, 'layer3': 3}

    # model = models.wide_resnet50_2(pretrained=True)  # 能力 21.49 5.91  ---top1
    # model = models.resnext50_32x4d(pretrained=True)  # 能力 22.38 6.30 ---top3
    # model = models.mobilenet_v2(pretrained=True)  # 能力 28.12 9.71 ---速度top1
    # model = models.mnasnet1_0(pretrained=True)
    # model = models.shufflenet_v2_x1_0(pretrained=True)

    # model = models.resnet50(pretrained=True)  # 下采样倍数32 能力23.85 7.13
    model = models.resnet34(pretrained=True)  # 下采样倍数32 能力23.85 7.13
    # return_layers = {'layer2': 1, 'layer3': 2, 'layer4': 3}
    # my_model = FModelOne2More(model, return_layers)
    # data_outputs = my_model(data_inputs_ts)
    # for k, v in data_outputs.items():
    #     print(v.shape)

    f_calc_flops_params(model, (300, 300))

    # f替换(model)

    # f_look_tw(model, data_inputs_list, name='f_look_tw')
    # f_look_summary(model, data_inputs_list)

    # model = darknet53()
    # modelsize(model, torch.rand(1, 3, 416, 416))
    '''
    torch.Size([1, 512, 80, 80])
    torch.Size([1, 1024, 40, 40])
    torch.Size([1, 2048, 20, 20])
    '''

    # model = MobileNetV1()  # 下采样倍数32 能力23.85 7.13
    # model = models.squeezenet1_0(pretrained=True)
    # model = models.vgg.vgg16(pretrained=True)
    # model = models.shufflenet_v2_x1_0(pretrained=True)
    # model = models.googlenet(pretrained=True)
    # model = models.mnasnet1_0(pretrained=True)  # 能力 26.49 8.456
    # model = models.inception_v3(pretrained=True)  # 能力 22.55 6.44

    '''-----------------模型分析 开始-----------------------'''

    # 用这个即可---查看网络的统计结果---
    # args_pd = tw.model_stats(model, data_inputs_list)
    # args_pd.to_excel('model_log.xlsx')

    # # print(type(args_pd))
    # print(args_pd)

    # from torchsummary import summary

    # summary1 = summary(model, (3, 640, 640))
    # print(type(summary1))

    # other()
    '''-----------------模型分析 完成-----------------------'''

    print('---%s--main执行完成------ ', os.path.basename(__file__))



import math
import os

import torch
import cv2
import numpy as np
import types
from numpy import random

from tools.GLOBAL_LOG import flog


class BasePretreatment:

    def __init__(self, cfg=None) -> None:
        self.cfg = cfg


def cre_transform_np4train(cfg):
    if cfg.tcfg_pic_handler is not None and (
            cfg.tcfg_pic_handler == 'widerface'
            or cfg.tcfg_pic_handler == 'face98'
    ):
        flog.warning('特殊数据处理 不要缩小 %s', )
        data_transform = {
            "train": Compose([
                Uint2Float32_np(),  # image int8转换成float [0,256)
                PhotometricDistort(),  # 图片处理集合
                # Expand(cfg.PIC_MEAN, cfg.NUM_KEYPOINTS),  # 放大缩小图片 只会缩小
                FRandomSampleCrop(cfg.NUM_KEYPOINTS),  # 随机剪切定位 keypoints 只会放大
                FRandomMirror(cfg.NAME_DATA),
                FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),
                Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
                ConvertColor(current='BGR', transform='RGB'),
                img_np_bgr2ts_rgb(is_box_oned=False),
            ], cfg)
        }
    elif cfg.MODE_VIS == 'keypoints':
        data_transform = {
            "train": Compose([
                Uint2Float32_np(),  # image int8转换成float [0,256)
                PhotometricDistort(),  # 图片处理集合
                FExpand(cfg.PIC_MEAN, cfg.NUM_KEYPOINTS),  # 放大缩小图片 只会缩小
                FRandomSampleCrop(cfg.NUM_KEYPOINTS),  # 随机剪切定位 keypoints 只会放大
                FRandomMirror(cfg.NAME_DATA),
                FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),
                Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
                ConvertColor(current='BGR', transform='RGB'),
                img_np_bgr2ts_rgb(is_box_oned=False),
            ], cfg)
        }
    else:
        if cfg.USE_BASE4NP:
            flog.error('使用的是 USE_BASE4NP 模式 %s', cfg.USE_BASE4NP)
            data_transform = {
                "train": Compose([
                    Uint2Float32_np(),  # image int8转换成float [0,256)
                    FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),
                    Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
                    ConvertColor(current='BGR', transform='RGB'),
                    img_np_bgr2ts_rgb(is_box_oned=False),
                ], cfg)
            }
        else:
            data_transform = {
                "train": Compose([
                    Uint2Float32_np(),  # image int8转换成float [0,256)
                    # ToAbsoluteCoords(),  # 输入已是原图不需要恢复 boxes 恢复原图尺寸
                    PhotometricDistort(),  # 图片处理集合
                    FExpand(cfg.PIC_MEAN),  # 放大缩小图片 无需cfg.NUM_KEYPOINTS
                    FRandomSampleCrop(),  # 随机剪切定位 无需cfg.NUM_KEYPOINTS
                    FRandomMirror(),  # 无需cfg.NAME_DATA
                    # ToPercentCoords(),  # boxes 按原图归一化 最后统一归一 最后ToTensor 统一归一
                    FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),
                    Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
                    ConvertColor(current='BGR', transform='RGB'),
                    img_np_bgr2ts_rgb(is_box_oned=False),
                ], cfg)
            }

    data_transform["val"] = Compose([
        FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),
        Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
        ConvertColor(current='BGR', transform='RGB'),
        img_np_bgr2ts_rgb(is_box_oned=False),
    ], cfg)

    return data_transform


def cre_transform_balance_data(cfg):
    data_transform = {
        "train": Compose([
            Uint2Float32_np(),  # image int8转换成float [0,256)
            # --------------- 图形整体 ------------
            RandomContrast(),  # 随机透明度
            ConvertColor(transform='HSV'),  # bgr -> hsv
            RandomSaturation(),  # 随机色彩'
            RandomHue(),  # HUE变化
            ConvertColor(current='HSV', transform='BGR'),  # hsv -> bgr
            RandomContrast(),  # 随机透明度
            RandomBrightness(),  # 随机亮度增强
            RandomGray(),  # 灰度
            FRandomNoise(),  # 杂点
            Ffilter_gaussian(),  # 高斯模糊
            Fgamma_transform(),  # Gamma 暗加强

            # --------------- 变换 ------------
            FExpand(cfg.PIC_MEAN, cfg.NUM_KEYPOINTS),  # 放大缩小图片 只会缩小
            FRandomSampleCrop(cfg.NUM_KEYPOINTS),  # 随机剪切定位 keypoints 只会放大
            FRandomMirror(cfg.NAME_DATA),  # 水平镜像
            FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),

            # --------------- 后处理 ------------
            Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
            ConvertColor(current='BGR', transform='RGB'),
            img_np_bgr2ts_rgb(is_box_oned=False),
        ], cfg)
    }
    data_transform["val"] = Compose([
        FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),
        Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
        ConvertColor(current='BGR', transform='RGB'),
        img_np_bgr2ts_rgb(is_box_oned=False),
    ], cfg)

    return data_transform


def cre_transform_np4test(cfg):
    data_transform = Compose([
        FPResize(cfg.P_SIZE, cfg.IS_SIZE_SCOPE),  # 针对一张图片
        Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
        ConvertColor(current='BGR', transform='RGB'),
        img_np_bgr2ts_rgb(is_box_oned=False),
    ], cfg)
    return data_transform


def cre_transform_np4base(cfg):
    flog.warning('预处理使用 cre_transform_base4np', )
    data_transform = {
        "train": Compose([
            FResize(cfg.IMAGE_SIZE, cfg.IS_MULTI_SCALE_V2, cfg.MULTI_SCALE_VAL_V2),
            Normalize(),
            ConvertColor(current='BGR', transform='RGB'),
            img_np_bgr2ts_rgb(is_box_oned=False),
        ], cfg)
    }

    # data_transform["val"] = Compose([
    #     FPResize(cfg.P_SIZE, cfg.P_SIZE),
    #     Normalize(cfg.PIC_MEAN, cfg.PIC_STD),
    #     ConvertColor(current='BGR', transform='RGB'),
    #     img_np_bgr2ts_rgb(is_box_oned=False),
    # ], cfg)

    return data_transform


def _copy_box(boxes):
    if isinstance(boxes, np.ndarray):
        boxes_ = boxes.copy()
    elif isinstance(boxes, torch.Tensor):
        boxes_ = boxes.clone()
    else:
        raise Exception('类型错误', type(boxes))

    return boxes_


def intersect(box_a, box_b):
    max_xy = np.minimum(box_a[:, 2:], box_b[2:])
    min_xy = np.maximum(box_a[:, :2], box_b[:2])
    inter = np.clip((max_xy - min_xy), a_min=0, a_max=np.inf)
    return inter[:, 0] * inter[:, 1]


def jaccard_numpy(box_a, box_b):
    """Compute the jaccard overlap of two sets of boxes.  The jaccard overlap
    is simply the intersection over union of two boxes.
    E.g.:
        A ∩ B / A ∪ B = A ∩ B / (area(A) + area(B) - A ∩ B)
    Args:
        box_a: Multiple bounding boxes, Shape: [num_boxes,4]
        box_b: Single bounding box, Shape: [4]
    Return:
        jaccard overlap: Shape: [box_a.shape[0], box_a.shape[1]]
    """
    inter = intersect(box_a, box_b)
    area_a = ((box_a[:, 2] - box_a[:, 0]) *
              (box_a[:, 3] - box_a[:, 1]))  # [A,B]
    area_b = ((box_b[2] - box_b[0]) *
              (box_b[3] - box_b[1]))  # [A,B]
    union = area_a + area_b - inter
    return inter / union  # [A,B]


class Compose(BasePretreatment):

    def __init__(self, transforms, cfg=None):
        super(Compose, self).__init__(cfg)
        self.transforms = transforms

    def __call__(self, img, target):
        # f_plt_show_cv(image,boxes)
        for t in self.transforms:
            img, target = t(img, target)
            if target is not None:
                if 'boxes' in target and 'labels' in target:
                    if len(target['boxes']) != len(target['labels']):
                        flog.warning('!!! 数据有问题 Compose  %s %s %s ', len(target['boxes']), len(target['labels']), t)
        return img, target


class Uint2Float32_np(object):
    def __call__(self, image, target):
        '''cv打开的np 默认是uint8'''
        return image.astype(np.float32), target


class ts2img_np_bgr(object):
    def __call__(self, tensor, target):
        return tensor.cpu().numpy().astype(np.float32).transpose((1, 2, 0)), target


# --------------- 图形整体 ------------

class FRandomNoise:
    '''
    随机噪声 噪点
    :param image:
    :param threshold:
    :return:
    '''

    def __init__(self, threshold=32) -> None:
        super().__init__()
        self.threshold = threshold

    def __call__(self, img_np, target):
        if random.randint(2):
            noise = np.random.uniform(low=-1, high=1, size=img_np.shape)
            img_np = img_np + noise * self.threshold
            img_np = np.clip(img_np, 0, 255)

        return img_np, target


class Ffilter_gaussian:
    '''
    高斯滤波器是根据高斯函数的形状来选择权值的线性平滑滤波器，滤波器符合二维高斯分布
    模糊
    :param image:
    :param threshold:
    :return:
    '''

    def __init__(self, k_size=5) -> None:
        super().__init__()
        self.k_size = k_size

    def __call__(self, img_np, target):
        if random.randint(2):
            img_np = cv2.GaussianBlur(img_np, (self.k_size, self.k_size), 3)
        return img_np, target


class Fgamma_transform:
    '''
    Gamma变换就是用来图像增强，通过非线性变换提升了暗部细节
    :param image:
    :param threshold:
    :return:
    '''

    def __init__(self, gamma=1.6) -> None:
        super().__init__()
        self.gamma = gamma

    def __call__(self, img_np, target):
        if random.randint(2):
            max_value = np.max(img_np)
            min_value = np.min(img_np)
            value_l = max_value - min_value
            img_np = (img_np - min_value) / value_l
            img_np = np.power(img_np, self.gamma)
            img_np = img_np * 255
        return img_np, target


class RandomSaturation(object):
    '''随机色彩 需要HSV'''

    def __init__(self, lower=0.5, upper=1.5):
        self.lower = lower
        self.upper = upper
        assert self.upper >= self.lower, "contrast upper must be >= lower."
        assert self.lower >= 0, "contrast lower must be non-negative."

    def __call__(self, image, target):
        if random.randint(2):
            image[:, :, 1] *= random.uniform(self.lower, self.upper)

        return image, target


class RandomHue(object):
    def __init__(self, delta=18.0):
        assert delta >= 0.0 and delta <= 360.0
        self.delta = delta

    def __call__(self, image, target):
        if random.randint(2):
            image[:, :, 0] += random.uniform(-self.delta, self.delta)
            image[:, :, 0][image[:, :, 0] > 360.0] -= 360.0
            image[:, :, 0][image[:, :, 0] < 0.0] += 360.0
        return image, target


class RandomContrast(object):
    '''随机透明度'''

    def __init__(self, lower=0.5, upper=1.5):
        self.lower = lower
        self.upper = upper
        assert self.upper >= self.lower, "contrast upper must be >= lower."
        assert self.lower >= 0, "contrast lower must be non-negative."

    # expects float image
    def __call__(self, image, target):
        if random.randint(2):  # 50%
            alpha = random.uniform(self.lower, self.upper)
            image *= alpha
        # f_plt_show_cv(image)
        return image, target


class RandomGray:
    '''随机灰度'''

    def __call__(self, image, target):
        if random.random() > 0.9:
            _img = np.zeros(image.shape, dtype=np.uint8)
            gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
            _img[:, :, 0] = gray
            _img[:, :, 1] = gray
            _img[:, :, 2] = gray
            image = _img

        return image, target


class RandomBrightness(object):
    def __init__(self, delta=32):
        assert delta >= 0.0
        assert delta <= 255.0
        self.delta = delta

    def __call__(self, image, target):
        '''随机亮度增强'''
        if random.randint(2):  # 50%
            delta = random.uniform(-self.delta, self.delta)
            image += delta
        # f_plt_show_cv(image)
        return image, target


# --------------- 变换 ------------
class FRandomRotate(object):
    '''
    有框不能用旋转
    '''

    def __init__(self, max_degree=20):
        self.max_degree = max_degree

    def __call__(self, img_np, target):
        degree = random.uniform(-self.max_degree, self.max_degree)
        h, w, c = img_np.shape
        cx, cy = w / 2.0, h / 2.0

        ''' 图片处理 '''
        # mat rotate 1 center 2 angle 3 缩放系数
        matRotate = cv2.getRotationMatrix2D((cy, cx), degree, 1.0)
        img_np = cv2.warpAffine(img_np, matRotate, (h, w))

        if 'boxes' in target:
            boxes_ltrb_ts = target['boxes']
            a = -degree / 180.0 * math.pi
            # boxes = torch.from_numpy(boxes)
            new_boxes_ltrb_ts = torch.zeros_like(boxes_ltrb_ts)
            new_boxes_ltrb_ts[:, 0] = boxes_ltrb_ts[:, 1]
            new_boxes_ltrb_ts[:, 1] = boxes_ltrb_ts[:, 0]
            new_boxes_ltrb_ts[:, 2] = boxes_ltrb_ts[:, 3]
            new_boxes_ltrb_ts[:, 3] = boxes_ltrb_ts[:, 2]
            for i in range(boxes_ltrb_ts.shape[0]):
                ymin, xmin, ymax, xmax = new_boxes_ltrb_ts[i, :]
                xmin, ymin, xmax, ymax = float(xmin), float(ymin), float(xmax), float(ymax)
                x0, y0 = xmin, ymin
                x1, y1 = xmin, ymax
                x2, y2 = xmax, ymin
                x3, y3 = xmax, ymax
                z = torch.FloatTensor([[y0, x0], [y1, x1], [y2, x2], [y3, x3]])
                tp = torch.zeros_like(z)
                tp[:, 1] = (z[:, 1] - cx) * math.cos(a) - (z[:, 0] - cy) * math.sin(a) + cx
                tp[:, 0] = (z[:, 1] - cx) * math.sin(a) + (z[:, 0] - cy) * math.cos(a) + cy
                ymax, xmax = torch.max(tp, dim=0)[0]
                ymin, xmin = torch.min(tp, dim=0)[0]
                new_boxes_ltrb_ts[i] = torch.stack([ymin, xmin, ymax, xmax])
            new_boxes_ltrb_ts[:, 1::2].clamp_(min=0, max=h - 1)
            new_boxes_ltrb_ts[:, 0::2].clamp_(min=0, max=w - 1)
            boxes_ltrb_ts[:, 0] = new_boxes_ltrb_ts[:, 1]
            boxes_ltrb_ts[:, 1] = new_boxes_ltrb_ts[:, 0]
            boxes_ltrb_ts[:, 2] = new_boxes_ltrb_ts[:, 3]
            boxes_ltrb_ts[:, 3] = new_boxes_ltrb_ts[:, 2]
            target['boxes'] = boxes_ltrb_ts

        if 'keypoints' in target:
            raise Exception('关键点检测不在这里做')
        return img_np, target


class FExpand(object):
    '''随机 缩小放在图片中某处  其它部份为黑色'''

    def __init__(self, mean, num_keypoints=None):
        ''' 均值用于扩展边界 '''
        self.mean = mean
        self.num_keypoints = num_keypoints

    def __call__(self, image, target):
        if random.randint(2):
            return image, target

        height, width, depth = image.shape
        ratio = random.uniform(1, 3)  # 缩小 1~4的一个比例
        left = random.uniform(0, width * ratio - width)
        top = random.uniform(0, height * ratio - height)

        expand_image = np.zeros(
            (int(height * ratio), int(width * ratio), depth),
            dtype=image.dtype)
        expand_image[:, :, :] = self.mean
        expand_image[int(top):int(top + height),
        int(left):int(left + width)] = image
        image = expand_image
        # 原尺寸不要了
        boxes = _copy_box(target['boxes'])

        if isinstance(boxes, np.ndarray):
            boxes[:, :2] += (int(left), int(top))  # lt
            boxes[:, 2:] += (int(left), int(top))  # rb
        elif isinstance(boxes, torch.Tensor):
            boxes[:, :2] += torch.tensor((int(left), int(top)))
            boxes[:, 2:] += torch.tensor((int(left), int(top)))
        else:
            raise Exception('target[boxes]类型错误', type(boxes))

        if 'keypoints' in target:
            assert self.num_keypoints > 0, 'cfg.NUM_KEYPOINTS = %s 设置错误' % self.num_keypoints
            if isinstance(boxes, np.ndarray):
                target['keypoints'] += (int(left), int(top)) * self.num_keypoints
            elif isinstance(boxes, torch.Tensor):
                target['keypoints'] += torch.tensor((int(left), int(top)) * self.num_keypoints)
            else:
                raise Exception('target[keypoints]类型错误', type(target['keypoints']))

        target['boxes'] = boxes
        # f_plt_show_cv(image,torch.tensor(boxes))
        return image, target


class FRandomSampleCrop(object):
    """
    随机Crop
    """

    def __init__(self, num_keypoints=None):
        # 随机源
        self.sample_options = (
            # using entire original input image
            None,
            # sample a patch s.t. MIN jaccard w/ obj in .1,.3,.4,.7,.9
            (0.1, None),
            (0.3, None),
            (0.7, None),
            (0.9, None),
            # randomly sample a patch
            (None, None),
        )
        self.num_keypoints = num_keypoints

    def __call__(self, image, target):
        height, width, _ = image.shape
        while True:
            # randomly choose a mode
            mode = random.choice(self.sample_options)
            if mode is None:
                return image, target

            min_iou, max_iou = mode
            if min_iou is None:
                min_iou = float('-inf')
            if max_iou is None:
                max_iou = float('inf')

            # max trails (50)
            for _ in range(50):
                current_image = image

                w = random.uniform(0.3 * width, width)
                h = random.uniform(0.3 * height, height)

                # aspect ratio constraint b/t .5 & 2
                if h / w < 0.5 or h / w > 2:
                    continue

                left = random.uniform(width - w)
                top = random.uniform(height - h)

                # convert to integer rect x1,y1,x2,y2
                rect = np.array([int(left), int(top), int(left + w), int(top + h)])

                # calculate IoU (jaccard overlap) b/t the cropped and gt boxes
                overlap = jaccard_numpy(target['boxes'], rect)

                # is min and max overlap constraint satisfied? if not try again
                if overlap.min() < min_iou and max_iou < overlap.max():
                    continue

                # cut the crop from the image
                current_image = current_image[rect[1]:rect[3], rect[0]:rect[2], :]

                # keep overlap with gt box IF center in sampled patch
                centers = (target['boxes'][:, :2] + target['boxes'][:, 2:]) / 2.0

                # mask in all gt boxes that above and to the left of centers
                m1 = (rect[0] < centers[:, 0]) * (rect[1] < centers[:, 1])

                # mask in all gt boxes that under and to the right of centers
                m2 = (rect[2] > centers[:, 0]) * (rect[3] > centers[:, 1])

                # mask in that both m1 and m2 are true
                mask = m1 * m2

                # have any valid boxes? try again if not
                if not mask.any():
                    continue

                # take only matching gt boxes
                # current_boxes = boxes[mask, :].copy()
                current_boxes = _copy_box(target['boxes'][mask, :])

                if 'keypoints' in target:
                    assert self.num_keypoints > 0, 'cfg.NUM_KEYPOINTS = %s 设置错误' % self.num_keypoints
                    target['keypoints'] = target['keypoints'][mask, :]
                    target['kps_mask'] = target['kps_mask'][mask, :]
                    # ltrb
                    rect_lt_x5 = np.tile(rect[:2], self.num_keypoints).astype(np.float32)
                    target['keypoints'] -= rect_lt_x5
                    img_wh_kp_ts = torch.tensor(current_image.shape[:2][::-1]).repeat(self.num_keypoints)
                    fverify_keypoints_2d(target, img_wh_kp_ts=img_wh_kp_ts)

                # take only matching gt labels
                current_labels = target['labels'][mask]

                # should we use the box left and top corner or the crop's
                current_boxes[:, :2] = np.maximum(current_boxes[:, :2], rect[:2])
                # adjust to crop (by substracting crop's left,top)
                current_boxes[:, :2] -= rect[:2]

                current_boxes[:, 2:] = np.minimum(current_boxes[:, 2:], rect[2:])
                # adjust to crop (by substracting crop's left,top)
                current_boxes[:, 2:] -= rect[:2]

                target['labels'] = current_labels
                target['boxes'] = current_boxes

                return current_image, target


class FRandomMirror(object):
    '''随机水平镜像'''

    def __init__(self, data_name=None) -> None:
        super().__init__()
        self.name_data = data_name
        self.rules_flip = {
            'face98': {
                0: 32, 1: 31, 2: 30, 3: 29, 4: 28, 5: 27, 6: 26, 7: 25, 8: 24, 9: 23, 10: 22, 11: 21, 12: 20, 13: 19,
                14: 18,
                15: 17,
                16: 16, 17: 15, 18: 14, 19: 13, 20: 12, 21: 11, 22: 10, 23: 9, 24: 8, 25: 7, 26: 6, 27: 5, 28: 4, 29: 3,
                30: 2,
                31: 1, 32: 0,
                33: 46, 34: 45, 35: 44, 36: 43, 37: 42, 38: 50, 39: 49, 40: 48, 41: 47,
                46: 33, 45: 34, 44: 35, 43: 36, 42: 37, 50: 38, 49: 39, 48: 40, 47: 41,
                60: 72, 61: 71, 62: 70, 63: 69, 64: 68, 65: 75, 66: 74, 67: 73,
                72: 60, 71: 61, 70: 62, 69: 63, 68: 64, 75: 65, 74: 66, 73: 67,
                96: 97, 97: 96,
                51: 51, 52: 52, 53: 53, 54: 54,
                55: 59, 56: 58, 57: 57, 58: 56, 59: 55,
                76: 82, 77: 81, 78: 80, 79: 79, 80: 78, 81: 77, 82: 76,
                87: 83, 86: 84, 85: 85, 84: 86, 83: 87,
                88: 92, 89: 91, 90: 90, 91: 89, 92: 88,
                95: 93, 94: 94, 93: 95
            },
            'face5': {0: 1, -1: -1, 2: 2, 3: 4, -1: -1},  # 水平翻转 这个要和数据遍历对起
            'widerface': {0: 1, -1: -1, 2: 2, 3: 4, -1: -1},
        }

    def __call__(self, image, target):
        _, width, _ = image.shape
        if random.randint(2):
            image = image[:, ::-1]
            # boxes = boxes.copy()
            boxes = _copy_box(target['boxes'])
            # boxes[:, 0::2] = width - boxes[:, 2::-2]
            boxes[:, [0, 2]] = width - boxes[:, [2, 0]]
            target['boxes'] = boxes

            if 'keypoints' in target:
                # if self.data_name is None:
                #     raise Exception('RandomMirror keypoints 不支持 需传入 data_name')
                assert self.name_data is not None, 'RandomMirror keypoints 不支持 需传入 name_data cfg.NAME_DATA '
                # 批量修改x值 y不变
                target['keypoints'][:, ::2] = width - target['keypoints'][:, ::2]

                ''' 这个需要订制一个翻转规则  '''
                if self.name_data not in self.rules_flip:
                    raise Exception('self.name_data 不存在 %s' % self.name_data)
                rule_flip = self.rules_flip[self.name_data]

                # 交换数组位置
                n = target['keypoints'].shape[0]
                kks = target['keypoints'].view(n, -1, 2)
                mms = target['kps_mask'].view(n, -1, 2)

                for i, _ in enumerate(kks):
                    # 改变 kk可以改变 target['keypoints'] 共享内存
                    kk = kks[i]
                    mm = mms[i]
                    for k, v in rule_flip.items():
                        if k == -1:
                            continue
                        _t = kk[k].clone()
                        kk[k] = kk[v]
                        kk[v] = _t
                        _t = mm[k].clone()
                        mm[k] = mm[v]
                        mm[v] = _t
                # 共享内存直接调整 无需这个
                # target['keypoints'] = kks.reshape(n, -1)
                # target['kps_mask'] = mms.reshape(n, -1)

            # target['keypoints'].view(-1,2).flip(0)

        return image, target


class FResize(object):
    def __init__(self, size_wh=None, is_multi_scale=False, multi_scale_val=(800, 1333)):
        '''
        如果 不是多尺寸 则按size_wh
        '''
        self.size_wh = size_wh
        self.is_multi_scale = is_multi_scale
        self.multi_scale_val = multi_scale_val

    def __call__(self, img_np, target):
        if self.is_multi_scale:
            min_side, max_side = self.multi_scale_val
            h, w, _ = img_np.shape
            smallest_side = min(w, h)
            largest_side = max(w, h)
            scale = min_side / smallest_side
            if largest_side * scale > max_side:
                scale = max_side / largest_side
            nw, nh = int(scale * w), int(scale * h)
            image_resized = cv2.resize(img_np, (nw, nh))

            if target is not None:
                target['toone'] = image_resized.shape[:2][::-1]
            pad_w = 32 - nw % 32
            pad_h = 32 - nh % 32
            # 右下加上pad 不影响targets
            image_res = np.zeros(shape=[nh + pad_h, nw + pad_w, 3], dtype=np.uint8)
            image_res[:nh, :nw, :] = image_resized

            scale_w = scale
            scale_h = scale
        else:
            assert self.size_wh is not None
            size_wh_z = self.size_wh
            w_ratio, h_ratio = np.array(img_np.shape[:2][::-1]) / np.array(size_wh_z)

            image_res = cv2.resize(img_np, (size_wh_z[0], size_wh_z[1]))

            scale_w = 1 / w_ratio
            scale_h = 1 / h_ratio

        if target is not None:
            if 'boxes' in target:
                target['boxes'][:, [0, 2]] = target['boxes'][:, [0, 2]] * scale_w
                target['boxes'][:, [1, 3]] = target['boxes'][:, [1, 3]] * scale_h
            if 'keypoints' in target:
                target['keypoints'][:, ::2] *= scale_w
                target['keypoints'][:, 1::2] *= scale_h

            # f_show_kp_np4plt(image_res, target['boxes'],
            #                  kps_xy_input=target['keypoints'],
            #                  mask_kps=target['kps_mask'],
            #                  is_recover_size=False)
            # f_show_od_pil4plt(image_res, target['boxes'], is_recover_size=False)
        return image_res, target


# --------------- 后处理 ------------
class Normalize(object):
    # def __init__(self, mean=(0.406, 0.456, 0.485), std=(0.225, 0.224, 0.229)):
    def __init__(self, mean=(128.0, 128.0, 128.0), std=(256.0, 256.0, 256.0)):
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)

    def __call__(self, image, target):
        ''' -0.5 ~ 0.5 '''
        # bgr
        # f_show_pic_np4plt(image)

        # (928, 1216, 3)
        image = image.astype(np.float32).clip(0.0, 255.)
        # image /= 255.
        image -= self.mean.reshape(1, 1, -1)
        image /= self.std

        if image.max() > 0.5 or image.min() < -0.5:
            raise Exception('图片有问题 mean = %s ,std = %s, max=%s,min=%s'
                            % (str(self.mean), str(self.std), image.max(), image.min()))

        # f_show_pic_np4plt(image)
        # image = f_recover_normalization4np(image)
        # f_show_pic_np4plt(image)

        return image, target


class ConvertColor(object):
    '''bgr -> hsv'''

    def __init__(self, current='BGR', transform='HSV'):
        self.transform = transform
        self.current = current

    def __call__(self, image, target):
        # plt.imshow(image)
        # plt.show()
        if self.current == 'BGR' and self.transform == 'HSV':
            image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        elif self.current == 'HSV' and self.transform == 'BGR':
            image = cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
        elif self.current == 'BGR' and self.transform == 'RGB':
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        else:
            raise NotImplementedError
        return image, target


class img_np_bgr2ts_rgb(object):
    def __init__(self, is_box_oned=False) -> None:
        super().__init__()
        self.is_box_oned = is_box_oned

    def __call__(self, cvimage, target):
        if target and self.is_box_oned:
            # np整体复制 wh
            whwh = np.tile(cvimage.shape[:2][::-1], 2)
            target['boxes'][:, :] = target['boxes'][:, :] / whwh
            if 'keypoints' in target:
                target['keypoints'][:, ::2] /= whwh[0]
                target['keypoints'][:, 1::2] /= whwh[1]

            # plt.imshow(cvimage)
            # plt.show()
        # (h,w,c -> c,h,w) = bgr
        return torch.from_numpy(cvimage.astype(np.float32)).permute(2, 0, 1), target


# ------------------------------ 无用 --------------------------------
class Target2one(object):
    '''原图转归一化尺寸  ToAbsoluteCoords 相反'''

    def __call__(self, image, target):
        height, width, channels = image.shape
        if target is not None:
            if 'boxes' in target:
                target['boxes'][:, 0] /= width
                target['boxes'][:, 2] /= width
                target['boxes'][:, 1] /= height
                target['boxes'][:, 3] /= height

            if 'keypoints' in target:
                target['keypoints'][:, ::2] /= width
                target['keypoints'][:, 1::2] /= height

        return image, target


class ToAbsoluteCoords(object):
    ''' boxes 恢复原图尺寸  归一化尺寸转原图  ToAbsoluteCoords 相反'''

    def __call__(self, image, target):
        '''
        归一化 -> 绝对坐标
        :param image:
        :param boxes:
        :param labels:
        :return:
        '''
        height, width, channels = image.shape
        if target is not None:
            if 'boxes' in target:
                # 这里可以直接改
                target['boxes'][:, 0] *= width
                target['boxes'][:, 2] *= width
                target['boxes'][:, 1] *= height
                target['boxes'][:, 3] *= height
            if 'keypoints' in target:
                target['keypoints'][:, ::2] *= width
                target['keypoints'][:, 1::2] *= height

        return image, target


class FPResize(object):
    def __init__(self, size, is_size_scope=False):
        '''
        size 可以是尺寸和范围 要求wh
        '''
        self.size = size
        self.is_size_scope = is_size_scope

    def __call__(self, img_np, target):
        if self.is_size_scope:
            min_side, max_side = self.size
            h, w, _ = img_np.shape
            smallest_side = min(w, h)
            largest_side = max(w, h)
            scale = min_side / smallest_side
            if largest_side * scale > max_side:
                scale = max_side / largest_side
            nw, nh = int(scale * w), int(scale * h)
            image_res = cv2.resize(img_np, (nw, nh))

            if target is not None:
                target['toone'] = image_res.shape[:2][::-1]

            scale_w = scale
            scale_h = scale
        else:
            w_ratio, h_ratio = np.array(img_np.shape[:2][::-1]) / np.array(self.size)

            image_res = cv2.resize(img_np, (self.size[0], self.size[1]))

            scale_w = 1 / w_ratio
            scale_h = 1 / h_ratio

        if target is not None:
            if 'boxes' in target:
                target['boxes'][:, [0, 2]] = target['boxes'][:, [0, 2]] * scale_w
                target['boxes'][:, [1, 3]] = target['boxes'][:, [1, 3]] * scale_h
            if 'keypoints' in target:
                target['keypoints'][:, ::2] *= scale_w
                target['keypoints'][:, 1::2] *= scale_h

        return image_res, target


class Lambda(object):
    """Applies a lambda as a transform. 这个用于复写方法?"""

    def __init__(self, lambd):
        assert isinstance(lambd, types.LambdaType)
        self.lambd = lambd

    def __call__(self, img, target):
        return self.lambd(img, target)


class PhotometricDistort(object):
    '''图片增强'''

    def __init__(self):
        self.pd = [
            RandomContrast(),  # 随机透明度
            ConvertColor(transform='HSV'),  # bgr -> hsv
            RandomSaturation(),  # 随机色彩'
            RandomHue(),
            ConvertColor(current='HSV', transform='BGR'),  # hsv -> bgr
            RandomContrast()  # 随机透明度
        ]
        self.rand_brightness = RandomBrightness()  # 随机亮度增强
        # self.rand_light_noise = RandomLightingNoise()  # 颜色杂音

    def __call__(self, image, target):
        im = image.copy()
        im, target = self.rand_brightness(im, target)
        if random.randint(2):  # 先转换还是后转换
            distort = Compose(self.pd[:-1])
        else:
            distort = Compose(self.pd[1:])
        im, target = distort(im, target)
        return im, target
    # return self.rand_light_noise(im, boxes, labels)


class SwapChannels(object):
    """
    随机 RGB 打乱
    Transforms a tensorized image by swapping the channels in the order
     specified in the swap tuple.
    Args:
        swaps (int triple): final order of channels
            eg: (2, 1, 0)
    """

    def __init__(self, swaps=(2, 1, 0)):
        self.swaps = swaps

    def __call__(self, image, target):
        """
        Args:
            image (Tensor): image tensor to be transformed
        Return:
            a tensor with channels swapped according to swap
        """
        # if torch.is_tensor(image):
        #     image = image.data.cpu().numpy()
        # else:
        #     image = np.array(image)
        image = image[:, :, self.swaps]
        return image, target


class KeyPointCropOne:
    '''
    支持 KeyPoint 随机剪图 + 归一化 图片没有归一化
    用这个 ToTensor is_box_oned=False
    '''

    def __init__(self, size_f, keep_scale_wh=False):
        self.keep_scale_wh = keep_scale_wh  # 这个标签不支持不能用
        self.size_f = size_f
        self.flip_landmarks_dict = {
            0: 32, 1: 31, 2: 30, 3: 29, 4: 28, 5: 27, 6: 26, 7: 25, 8: 24, 9: 23, 10: 22, 11: 21, 12: 20, 13: 19,
            14: 18,
            15: 17,
            16: 16, 17: 15, 18: 14, 19: 13, 20: 12, 21: 11, 22: 10, 23: 9, 24: 8, 25: 7, 26: 6, 27: 5, 28: 4, 29: 3,
            30: 2,
            31: 1, 32: 0,
            33: 46, 34: 45, 35: 44, 36: 43, 37: 42, 38: 50, 39: 49, 40: 48, 41: 47,
            46: 33, 45: 34, 44: 35, 43: 36, 42: 37, 50: 38, 49: 39, 48: 40, 47: 41,
            60: 72, 61: 71, 62: 70, 63: 69, 64: 68, 65: 75, 66: 74, 67: 73,
            72: 60, 71: 61, 70: 62, 69: 63, 68: 64, 75: 65, 74: 66, 73: 67,
            96: 97, 97: 96,
            51: 51, 52: 52, 53: 53, 54: 54,  # 这里不用换 79 90 94 等
            55: 59, 56: 58, 57: 57, 58: 56, 59: 55,
            76: 82, 77: 81, 78: 80, 79: 79, 80: 78, 81: 77, 82: 76,
            87: 83, 86: 84, 85: 85, 84: 86, 83: 87,
            88: 92, 89: 91, 90: 90, 91: 89, 92: 88,
            95: 93, 94: 94, 93: 95
        }

    def letterbox(self, img_, img_size=256, mean_rgb=(128, 128, 128)):
        shape_ = img_.shape[:2]  # shape = [height, width]
        ratio = float(img_size) / max(shape_)  # ratio  = old / new
        new_shape_ = (round(shape_[1] * ratio), round(shape_[0] * ratio))
        dw_ = (img_size - new_shape_[0]) / 2  # width padding
        dh_ = (img_size - new_shape_[1]) / 2  # height padding
        top_, bottom_ = round(dh_ - 0.1), round(dh_ + 0.1)
        left_, right_ = round(dw_ - 0.1), round(dw_ + 0.1)
        # resize img
        img_a = cv2.resize(img_, new_shape_, interpolation=cv2.INTER_LINEAR)

        img_a = cv2.copyMakeBorder(img_a, top_, bottom_, left_, right_, cv2.BORDER_CONSTANT,
                                   value=mean_rgb)  # padded square

        return img_a

    def __call__(self, image, target, site_xy, vis=False):
        '''

        :param image:
        :param target:
        :param site_xy: 旋转锚点
        :param vis:
        :return:
        '''
        cx, cy = site_xy
        pts = target['keypoints']
        angle = random.randint(-36, 36)

        (h, w) = image.shape[:2]
        h = h
        w = w
        # (cx , cy) = (int(0.5 * w) , int(0.5 * h))
        M = cv2.getRotationMatrix2D((cx, cy), angle, 1.0)
        cos = np.abs(M[0, 0])
        sin = np.abs(M[0, 1])

        # 计算新图像的bounding
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))

        M[0, 2] += int(0.5 * nW) - cx
        M[1, 2] += int(0.5 * nH) - cy

        resize_model = [cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_NEAREST, cv2.INTER_AREA, cv2.INTER_LANCZOS4]

        img_rot = cv2.warpAffine(image, M, (nW, nH), flags=resize_model[random.randint(0, 4)])
        # flags : INTER_LINEAR INTER_CUBIC INTER_NEAREST
        # borderMode : BORDER_REFLECT BORDER_TRANSPARENT BORDER_REPLICATE CV_BORDER_WRAP BORDER_CONSTANT

        pts_r = []
        for pt in pts:
            x = float(pt[0])
            y = float(pt[1])

            x_r = (x * M[0][0] + y * M[0][1] + M[0][2])
            y_r = (x * M[1][0] + y * M[1][1] + M[1][2])

            pts_r.append([x_r, y_r])

        x = [pt[0] for pt in pts_r]
        y = [pt[1] for pt in pts_r]

        x1, y1, x2, y2 = np.min(x), np.min(y), np.max(x), np.max(y)

        translation_pixels = 60

        scaling = 0.3
        x1 += random.randint(-int(max((x2 - x1) * scaling, translation_pixels)), int((x2 - x1) * 0.25))
        y1 += random.randint(-int(max((y2 - y1) * scaling, translation_pixels)), int((y2 - y1) * 0.25))
        x2 += random.randint(-int((x2 - x1) * 0.15), int(max((x2 - x1) * scaling, translation_pixels)))
        y2 += random.randint(-int((y2 - y1) * 0.15), int(max((y2 - y1) * scaling, translation_pixels)))

        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        x1 = int(max(0, x1))
        y1 = int(max(0, y1))
        x2 = int(min(x2, img_rot.shape[1] - 1))
        y2 = int(min(y2, img_rot.shape[0] - 1))

        crop_rot = img_rot[y1:y2, x1:x2, :]

        crop_pts = []
        width_crop = float(x2 - x1)
        height_crop = float(y2 - y1)
        # 归一化
        for pt in pts_r:
            x = pt[0]
            y = pt[1]
            crop_pts.append([float(x - x1) / width_crop, float(y - y1) / height_crop])

        # 随机镜像 这个存在标签左右眼切换问题 98点可以支持 其它需定制
        if random.random() >= 0.5:
            # print('--------->>> flip')
            crop_rot = cv2.flip(crop_rot, 1)
            crop_pts_flip = []
            for i in range(len(crop_pts)):
                # print( crop_rot.shape[1],crop_pts[flip_landmarks_dict[i]][0])
                x = 1. - crop_pts[self.flip_landmarks_dict[i]][0]
                y = crop_pts[self.flip_landmarks_dict[i]][1]
                # print(i,x,y)
                crop_pts_flip.append([x, y])
            crop_pts = crop_pts_flip

        if vis:
            for pt in crop_pts:
                x = int(pt[0] * width_crop)
                y = int(pt[1] * height_crop)

                cv2.circle(crop_rot, (int(x), int(y)), 2, (255, 0, 255), -1)
            # cv2.imshow('img', crop_rot)
            from matplotlib import pyplot as plt
            plt.imshow(crop_rot)
            plt.show()

        if self.keep_scale_wh:
            # 这个标签不支持不能用
            # crop_rot = self.letterbox(crop_rot, img_size=self.size_f[0], mean_rgb=(128, 128, 128))
            raise Exception('self.keep_scale_wh = %s 不支持 ' % self.keep_scale_wh)
        else:
            crop_rot = cv2.resize(crop_rot, self.size_f, interpolation=resize_model[random.randint(0, 4)])

        if vis:
            for pt in crop_pts:
                x = int(pt[0] * self.size_f[0])
                y = int(pt[1] * self.size_f[1])

                cv2.circle(crop_rot, (int(x), int(y)), 2, (255, 0, 255), -1)
            # cv2.imshow('img', crop_rot)
            from matplotlib import pyplot as plt
            plt.imshow(crop_rot)
            plt.show()
        ''' 图片没有归一化 '''
        return image, target


def f_recover_normalization4ts(img_ts, mean_bgr=(128.0, 128.0, 128.0), std_bgr=(256.0, 256.0, 256.0)):
    '''

    :param img_ts: c,h,w
    :return:
    '''
    device = img_ts.device
    # torch.Size([3, 928, 1536]) -> torch.Size([928, 1536, 3])
    # img_ts_show = img_ts.permute(1, 2, 0) # rgb -> bgr
    mean_bgr = torch.tensor(mean_bgr, device=device)[:, None, None]  # 3,1,1
    std_bgr = torch.tensor(std_bgr, device=device).unsqueeze(-1).unsqueeze(-1)
    img_ts = img_ts * std_bgr + mean_bgr
    # img_ts_show = img_ts_show.permute(2, 0, 1)# bgr -> rgb
    return img_ts


def f_recover_normalization4np(img_np_bgr, mean_bgr=(128.0, 128.0, 128.0), std_bgr=(256.0, 256.0, 256.0)):
    '''

    :param img_np_bgr: c,h,w
    :return:
    '''
    mean_bgr = np.array(mean_bgr)
    std_bgr = np.array(std_bgr)
    img_np_bgr = img_np_bgr * std_bgr + mean_bgr
    return img_np_bgr


def random_rotation4ts(img_np, boxes_ltrb_ts, kps_3d_ts, degree=10):
    '''
    kps_ts: torch.Size([1, 5, 3])
    '''
    angle = random.uniform(-degree, degree)
    h, w, c = img_np.shape
    cx, cy = w / 2.0, h / 2.0

    ''' 图片处理 '''
    # mat rotate 1 center 2 angle 3 缩放系数
    matRotate = cv2.getRotationMatrix2D((cy, cx), angle, 1.0)
    img_np = cv2.warpAffine(img_np, matRotate, (h, w))

    # debug
    # print(angle)
    # plt.imshow(img_np)
    # plt.show()

    ''' box处理 '''
    a = -angle / 180.0 * math.pi
    # boxes = torch.from_numpy(boxes)
    new_boxes_ltrb_ts = torch.zeros_like(boxes_ltrb_ts)
    new_boxes_ltrb_ts[:, 0] = boxes_ltrb_ts[:, 1]
    new_boxes_ltrb_ts[:, 1] = boxes_ltrb_ts[:, 0]
    new_boxes_ltrb_ts[:, 2] = boxes_ltrb_ts[:, 3]
    new_boxes_ltrb_ts[:, 3] = boxes_ltrb_ts[:, 2]
    for i in range(boxes_ltrb_ts.shape[0]):
        ymin, xmin, ymax, xmax = new_boxes_ltrb_ts[i, :]
        xmin, ymin, xmax, ymax = float(xmin), float(ymin), float(xmax), float(ymax)
        x0, y0 = xmin, ymin
        x1, y1 = xmin, ymax
        x2, y2 = xmax, ymin
        x3, y3 = xmax, ymax
        z = torch.FloatTensor([[y0, x0], [y1, x1], [y2, x2], [y3, x3]])
        tp = torch.zeros_like(z)
        tp[:, 1] = (z[:, 1] - cx) * math.cos(a) - (z[:, 0] - cy) * math.sin(a) + cx
        tp[:, 0] = (z[:, 1] - cx) * math.sin(a) + (z[:, 0] - cy) * math.cos(a) + cy
        ymax, xmax = torch.max(tp, dim=0)[0]
        ymin, xmin = torch.min(tp, dim=0)[0]
        new_boxes_ltrb_ts[i] = torch.stack([ymin, xmin, ymax, xmax])
    new_boxes_ltrb_ts[:, 1::2].clamp_(min=0, max=h - 1)
    new_boxes_ltrb_ts[:, 0::2].clamp_(min=0, max=w - 1)
    boxes_ltrb_ts[:, 0] = new_boxes_ltrb_ts[:, 1]
    boxes_ltrb_ts[:, 1] = new_boxes_ltrb_ts[:, 0]
    boxes_ltrb_ts[:, 2] = new_boxes_ltrb_ts[:, 3]
    boxes_ltrb_ts[:, 3] = new_boxes_ltrb_ts[:, 2]
    # boxes_ts = boxes_ts.numpy()

    ngt, nkey, c = kps_3d_ts.shape
    for i in range(ngt):
        for j in range(nkey):
            x = kps_3d_ts[i][j][0]
            y = kps_3d_ts[i][j][1]
            p = np.array([x, y, 1])
            p = matRotate.dot(p)
            kps_3d_ts[i][j][0] = p[0]
            kps_3d_ts[i][j][1] = p[1]

    return img_np, boxes_ltrb_ts, kps_3d_ts


def random_rotation4np(img_np, boxes_ltrb_np, kps_3d_np, degree=10):
    angle = random.uniform(-degree, degree)
    h, w, c = img_np.shape
    cx, cy = w / 2.0, h / 2.0

    ''' 图片处理 '''
    # mat rotate 1 center 2 angle 3 缩放系数
    matRotate = cv2.getRotationMatrix2D((cy, cx), angle, 1.0)
    img_np = cv2.warpAffine(img_np, matRotate, (h, w))

    # debug
    # print(angle)
    # plt.imshow(img_np)
    # plt.show()

    ''' box处理 '''
    a = -angle / 180.0 * math.pi
    # boxes = torch.from_numpy(boxes)
    new_boxes_ltrb_np = np.zeros_like(boxes_ltrb_np)
    new_boxes_ltrb_np[:, 0] = boxes_ltrb_np[:, 1]
    new_boxes_ltrb_np[:, 1] = boxes_ltrb_np[:, 0]
    new_boxes_ltrb_np[:, 2] = boxes_ltrb_np[:, 3]
    new_boxes_ltrb_np[:, 3] = boxes_ltrb_np[:, 2]
    for i in range(boxes_ltrb_np.shape[0]):
        ymin, xmin, ymax, xmax = new_boxes_ltrb_np[i, :]
        xmin, ymin, xmax, ymax = float(xmin), float(ymin), float(xmax), float(ymax)
        x0, y0 = xmin, ymin
        x1, y1 = xmin, ymax
        x2, y2 = xmax, ymin
        x3, y3 = xmax, ymax
        z = np.array([[y0, x0], [y1, x1], [y2, x2], [y3, x3]], dtype=np.float32)
        tp = np.zeros_like(z)
        tp[:, 1] = (z[:, 1] - cx) * math.cos(a) - (z[:, 0] - cy) * math.sin(a) + cx
        tp[:, 0] = (z[:, 1] - cx) * math.sin(a) + (z[:, 0] - cy) * math.cos(a) + cy
        ymax, xmax = np.max(tp, axis=0)
        ymin, xmin = np.min(tp, axis=0)
        new_boxes_ltrb_np[i] = np.stack([ymin, xmin, ymax, xmax])

    new_boxes_ltrb_np[:, 1::2] = np.clip(new_boxes_ltrb_np[:, 0::2], a_min=0, a_max=h - 1)
    new_boxes_ltrb_np[:, 1::2] = np.clip(new_boxes_ltrb_np[:, 0::2], a_min=0, a_max=w - 1)
    boxes_ltrb_np[:, 0] = new_boxes_ltrb_np[:, 1]
    boxes_ltrb_np[:, 1] = new_boxes_ltrb_np[:, 0]
    boxes_ltrb_np[:, 2] = new_boxes_ltrb_np[:, 3]
    boxes_ltrb_np[:, 3] = new_boxes_ltrb_np[:, 2]

    ngt, nkey, c = kps_3d_np.shape
    for i in range(ngt):
        for j in range(nkey):
            x = kps_3d_np[i][j][0]
            y = kps_3d_np[i][j][1]
            p = np.array([x, y, 1])
            p = matRotate.dot(p)
            kps_3d_np[i][j][0] = p[0]
            kps_3d_np[i][j][1] = p[1]

    return img_np, boxes_ltrb_np, kps_3d_np


if __name__ == '__main__':
    from f_tools.datas.f_coco.coco_dataset import CustomCocoDataset, fverify_keypoints_2d

    mode = 'keypoints'  # bbox segm keypoints caption


    # mode = 'bbox'  # bbox segm keypoints caption

    class cfg:
        pass


    cfg.IMAGE_SIZE = (448, 448)
    cfg.PIC_MEAN = (128, 128, 128.)
    cfg.PIC_STD = (256, 256, 256)
    cfg.USE_BASE4NP = False  # 基础处理
    cfg.IS_VISUAL_PRETREATMENT = False  # 用于dataset提取时测试
    cfg.NUM_KEYPOINTS = 5
    cfg.MODE_VIS = 'keypoints'
    cfg.NAME_DATA = 'face5'
    cfg.tcfg_pic_handler = ''
    cfg.IS_MULTI_SCALE_V2 = True
    cfg.MULTI_SCALE_VAL_V2 = [900, 2000]

    P_SIZE = [416, 416]
    IS_SIZE_SCOPE = False

    transform = None
    # transform = FExpand(cfg.PIC_MEAN, cfg.NUM_KEYPOINTS)
    # transform = SwapChannels()
    # transform = FRandomRotate(90)
    # transform = FRandomNoise()
    # transform = Ffilter_mean()
    # transform = Fgamma_transform()
    # transform = FRandomSampleCrop(cfg.NUM_KEYPOINTS)
    # transform = FResize([500, 100])
    # transform = FResize(is_multi_scale=True, multi_scale_val=[800, 1333])
    # transform = FPResize([300, 200])
    # transform = FPResize([100, 300], is_size_scope=True)
    # transform = RandomMirror()
    # transform = img_np_bgr2ts_rgb()

    # transform = cre_transform_np4train(cfg)['train']
    # transform = cre_transform_np4base(cfg)['train']
    # transform = cre_transform_balance_data(cfg)['train']
    # transform = Compose([
    #     # ConvertFromInts(),  # image int8转换成float [0,256)
    #     # ToAbsoluteCoords(), # 恢复真实尺寸
    #     # PhotometricDistort(),  # 图片处理集合
    #     FExpand(cfg.PIC_MEAN,cfg.NUM_KEYPOINTS),  # 放大缩小图片
    #     # FRandomSampleCrop(cfg.NUM_KEYPOINTS),  # 随机剪切定位
    #     # FRandomMirror('face5'),
    #     # ToPercentCoords(),  # boxes 按原图归一化 最后ToTensor 统一归一
    #     # Resize(cfg.IMAGE_SIZE),  # 定义模型输入尺寸 处理img和boxes
    #     Normalize(cfg.PIC_MEAN, cfg.PIC_STD),  # 正则化图片
    #     # ConvertColor(current='BGR', transform='RGB'),
    #     # ToTensor(is_box_oned=True),  # img 及 boxes(可选,前面已归一)归一  转tensor
    # ], cfg)

    # path_img, dataset = load_dataset_coco(mode, transform=transform)

    # path_root = r'M:\AI\datas\face_98'  # 自已的数据集
    # file_json = os.path.join(path_root, 'annotations', 'keypoints_test_2500_2118.json')
    # path_img = os.path.join(path_root, 'images_test_2118')

    # file_json = os.path.join(path_root, 'annotations', 'keypoints_train_7500_5316.json')
    # path_img = os.path.join(path_root, 'images_train_5316')

    # path_root = r'M:\AI\datas\face_5'
    path_root = r'/AI/datas/face_5'
    file_json = os.path.join(path_root, 'annotations', 'keypoints_train_10000_10000.json')
    path_img = os.path.join(path_root, 'images_13466')

    # path_root = r'/AI/datas/VOC2007'
    # file_json = os.path.join(path_root, 'coco/annotations', 'instances_type3_train_1066.json')
    # path_img = os.path.join(path_root, 'train/JPEGImages')

    dataset = CustomCocoDataset(
        file_json=file_json,
        path_img=path_img,
        mode=mode,
        transform=transform,
        cfg=cfg,
    )
    from f_tools.pic.f_show import f_show_od_np4plt_v1, fshow_kp_ts4plt, f_show_kp_np4plt, f_show_kp_np4plt_new

    for data in dataset:
        img_np_or_tensor, target = data

        if mode == 'keypoints':
            if isinstance(img_np_or_tensor, np.ndarray):
                f_show_kp_np4plt(img_np_or_tensor, target['boxes'],
                                 kps_xy_input=target['keypoints'],
                                 mask_kps=torch.ones_like(target['keypoints'], dtype=torch.bool),
                                 )

                kps_3d_ts = target['keypoints']
                b, c = kps_3d_ts.shape
                kps_3d_ts = kps_3d_ts.reshape(b, 5, 2)
                a1 = torch.ones(b, 5, 1)
                # torch.Size([1, 5, 3])
                kps_3d_ts = torch.cat([kps_3d_ts, a1], -1)
                kps_3d_np = kps_3d_ts.numpy()
                # img_np_or_tensor, target['boxes'], kps_ts \
                #     = random_rotation4ts(img_np_or_tensor, target['boxes'], kps_ts, 90)

                img_np_or_tensor, target['boxes'], kps_3d_np \
                    = random_rotation4np(img_np_or_tensor, target['boxes'].numpy(), kps_3d_np, 90)

                f_show_kp_np4plt_new(img_np_or_tensor, target['boxes'],
                                     kps_xy_input=kps_3d_ts,
                                     )

            else:
                fshow_kp_ts4plt(img_np_or_tensor,
                                target['boxes'],
                                target['keypoints'],
                                mask_kps=target['kps_mask'],
                                is_recover_size=True
                                )  # 需要恢复box
        else:
            # img_np_or_tensor = f_recover_normalization4ts(img_np_or_tensor)
            f_show_od_np4plt_v1(img_np_or_tensor, target['boxes'],
                                # g_ltrb=targets[i]['boxes'].cpu() / size_wh_toone_ts_x2 * size_wh_file_x2,
                                # 这里有错? GT默认不归一化
                                # ids2classes=data_loader.dataset.ids_classes,
                                # labels=p_labels[mask],
                                # scores=p_scores[mask].tolist(),
                                is_recover_size=False
                                )

        # img_np_or_tensor, target['boxes'] = random_rotation_np(img_np_or_tensor, target['boxes'])
        # f_show_od_ts4plt_v3(img_np_or_tensor, target['boxes'], is_recover_size=False)

        # f_show_kp_np4plt(img_np_or_tensor, target['boxes'],
        #                  kps_xy_input=target['keypoints'],
        #                  mask_kps=target['kps_mask'],
        #                  # 测试集 GT默认不归一化,是input模型尺寸
        #                  # g_boxes_ltrb=targets[i]['boxes'].cpu() / _size * wh_x2,
        #                  # ids2classes=data_loader.dataset.ids_classes,
        #                  # labels=p_labels[mask],
        #                  # scores=p_scores[mask].tolist(),
        #                  is_recover_size=False)

        # f_show_od_np4plt_v1(img_np_or_tensor, target['boxes'],
        #                     # g_ltrb=targets[i]['boxes'].cpu() / size_wh_toone_ts_x2 * size_wh_file_x2,
        #                     # 这里有错? GT默认不归一化
        #                     # ids2classes=data_loader.dataset.ids_classes,
        #                     # labels=p_labels[mask],
        #                     # scores=p_scores[mask].tolist(),
        #                     is_recover_size=False
        #                     )



import torch
from torch import nn
import torch.nn.functional as F


class SiLU(nn.Module):  # export-friendly version of nn.SiLU()
    @staticmethod
    def forward(x):
        return x * torch.sigmoid(x)


class SwishImplementation(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return x * torch.sigmoid(x)

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_tensors[0]
        sx = torch.sigmoid(x)  # sigmoid(ctx)
        return grad_output * (sx * (1 + x * (1 - sx)))


class MishImplementation(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return x.mul(torch.tanh(F.softplus(x)))  # x * tanh(ln(1 + exp(x)))

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_tensors[0]
        sx = torch.sigmoid(x)
        fx = F.softplus(x).tanh()
        return grad_output * (fx + x * sx * (1 - fx * fx))


class MemoryEfficientSwish(nn.Module):
    def forward(self, x):
        return SwishImplementation.apply(x)


class MemoryEfficientMish(nn.Module):
    def forward(self, x):
        return MishImplementation.apply(x)


class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class HardSwish(nn.Module):  # https://arxiv.org/pdf/1905.02244.pdf
    def forward(self, x):
        return x * F.hardtanh(x + 3, 0., 6., True) / 6.
        # return x * F.relu6(x + 3.0) / 6.0


class Mish(nn.Module):  # https://github.com/digantamisra98/Mish
    # LeakyReLU => Mish
    def forward(self, x):
        return x * F.softplus(x).tanh()


# FReLU https://arxiv.org/abs/2007.11824 -------------------------------------------------------------------------------
class FReLU(nn.Module):
    def __init__(self, c1, k=3):  # ch_in, kernel
        super().__init__()
        self.conv = nn.Conv2d(c1, c1, k, 1, 1, groups=c1, bias=False)
        self.bn = nn.BatchNorm2d(c1)

    def forward(self, x):
        return torch.max(x, self.bn(self.conv(x)))



import math

import torch
from torch import nn
import torch.nn.functional as F
import numpy as np

from tools.fmodels.model_subassembly import FConv2d


class Bottleneck(nn.Module):
    # Standard bottleneck
    def __init__(self, in_ch, dilation=1, e=0.5):
        super(Bottleneck, self).__init__()
        inter_ch = int(in_ch * e)
        self.branch = nn.Sequential(
            FConv2d(in_ch, inter_ch, k=1, act='relu', is_bias=False),
            FConv2d(inter_ch, inter_ch, k=3, act='relu', p=dilation, d=dilation, is_bias=False),
            FConv2d(inter_ch, in_ch, k=1, act='relu', is_bias=False)
        )

    def forward(self, x):
        return x + self.branch(x)


class DilateEncoder(nn.Module):
    """ DilateEncoder 与SPP类似加在C5后面 """

    def __init__(self, in_ch, out_ch, dilation_list=[2, 4, 6, 8]):
        super(DilateEncoder, self).__init__()
        self.projector = nn.Sequential(
            FConv2d(in_ch, out_ch, k=1, act='identity', is_bias=False),
            FConv2d(out_ch, out_ch, k=3, p=1, act='identity', is_bias=False)
        )
        encoders = []
        for d in dilation_list:
            encoders.append(Bottleneck(in_ch=out_ch, dilation=d))
        self.encoders = nn.Sequential(*encoders)

    def forward(self, x):
        x = self.projector(x)
        x = self.encoders(x)

        return x


class UpSample(nn.Module):
    def __init__(self, size=None, scale_factor=None, mode='nearest', align_corner=None):
        super(UpSample, self).__init__()
        self.size = size
        self.scale_factor = scale_factor
        self.mode = mode
        self.align_corner = align_corner

    def forward(self, x):
        return torch.nn.functional.interpolate(x, size=self.size, scale_factor=self.scale_factor,
                                               mode=self.mode, align_corners=self.align_corner)


class ResizeConv(nn.Module):
    def __init__(self, in_ch, out_ch, size=None, scale_factor=None, mode='nearest', align_corner=None):
        super(ResizeConv, self).__init__()
        self.upsample = UpSample(size=size, scale_factor=scale_factor, mode=mode, align_corner=align_corner)
        self.conv = FConv2d(in_ch, out_ch, k=1, act='relu', is_bias=False)

    def forward(self, x):
        x = self.conv(self.upsample(x))
        return x


class SPPv2(torch.nn.Module):
    '''
    同尺寸同维
    多池化混合 通道增加4倍 尺寸不变
    一个输入 由多个尺寸的核 池化后 再进行堆叠
    增加感受野，分离出最显著的上下文特征
    '''

    def __init__(self, seq='asc'):
        super(SPPv2, self).__init__()
        assert seq in ['desc', 'asc']
        self.seq = seq

    def __call__(self, x):
        x_1 = x
        x_2 = F.max_pool2d(input=x, kernel_size=5, stride=1, padding=2)
        # x_3 = F.max_pool2d(x, pool_size, 1, pool_size//2)
        x_3 = F.max_pool2d(x, 9, 1, 4)
        x_4 = F.max_pool2d(x, 13, 1, 6)
        if self.seq == 'desc':
            out = torch.cat([x_4, x_3, x_2, x_1], dim=1)
        else:
            out = torch.cat([x_1, x_2, x_3, x_4], dim=1)
        return out


class SPPv3(torch.nn.Module):
    '''
    SPPv2 加入通道 压缩转换 参数量更少
    '''

    def __init__(self, in_channels, seq='asc'):
        super(SPPv3, self).__init__()
        assert seq in ['desc', 'asc']
        self.seq = seq

        _t = in_channels // 2

        self.cbs1 = FConv2d(in_channels, _t, k=1, act='silu')
        self.cbs2 = FConv2d(_t * (3 + 1), in_channels, k=1, act='silu')

    def __call__(self, x):
        x = self.cbs1(x)
        x_1 = x
        x_2 = F.max_pool2d(input=x, kernel_size=5, stride=1, padding=2)
        x_3 = F.max_pool2d(x, 9, 1, 4)
        x_4 = F.max_pool2d(x, 13, 1, 6)
        if self.seq == 'desc':
            out = torch.cat([x_4, x_3, x_2, x_1], dim=1)
        else:
            out = torch.cat([x_1, x_2, x_3, x_4], dim=1)
        out = self.cbs2(out)
        return out


class SELayer(nn.Module):
    def __init__(self, channel, reduction=16):
        super(SELayer, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction, channel, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y.expand_as(x)


class ChannelAttention(nn.Module):
    '''
    CBAM
    通道注意力机制
    '''

    def __init__(self, in_planes, rotio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.sharedMLP = nn.Sequential(
            nn.Conv2d(in_planes, in_planes // rotio, 1, bias=False),
            nn.ReLU(),
            nn.Conv2d(in_planes // rotio, in_planes, 1, bias=False))
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avgout = self.sharedMLP(self.avg_pool(x))
        maxout = self.sharedMLP(self.max_pool(x))
        return self.sigmoid(avgout + maxout)


class SpatialAttention(nn.Module):
    '''
    CBAM
    空间注意力机制
    '''

    def __init__(self, kernel_size=7):
        super(SpatialAttention, self).__init__()
        assert kernel_size in (3, 7), "kernel size must be 3 or 7"
        padding = 3 if kernel_size == 7 else 1

        self.conv = nn.Conv2d(2, 1, kernel_size, padding=padding, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avgout = torch.mean(x, dim=1, keepdim=True)
        maxout, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avgout, maxout], dim=1)
        x = self.conv(x)
        return self.sigmoid(x)


class SAM(nn.Module):
    """ conv -> sigmoid * x -> 对原值进行缩放  Parallel CBAM from yolov4"""

    def __init__(self, in_ch):
        super(SAM, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, in_ch, 1),
            nn.Sigmoid()
        )

    def forward(self, x):
        """ Spatial Attention Module """
        x_attention = self.conv(x)

        return x * x_attention


class _Bottleneck(nn.Module):
    '''
    Standard bottleneck
    结构见 Bottleneck.jpg
    减小通道后 恢复
    '''

    def __init__(self, c1, c2, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, shortcut, groups, expansion
        super(_Bottleneck, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = FConv2d(c1, c_, k=1, is_bias=False)
        self.cv2 = FConv2d(c_, c2, k=3, p=1, g=g, is_bias=False)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class BottleneckCSP(nn.Module):
    '''
    结构见 BottleneckCSP.jpg
    BottleneckCSP-CSP瓶颈层
    CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks
    '''

    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super(BottleneckCSP, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = FConv2d(c1, c_, k=1, is_bias=False)
        self.cv2 = nn.Conv2d(c1, c_, kernel_size=1, bias=False)
        self.cv3 = nn.Conv2d(c_, c_, kernel_size=1, bias=False)
        self.cv4 = FConv2d(2 * c_, c2, k=1, is_bias=False)
        self.bn = nn.BatchNorm2d(2 * c_)  # applied to cat(cv2, cv3)
        self.act = nn.LeakyReLU(0.1, inplace=True)
        self.m = nn.Sequential(*[_Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)])

    def forward(self, x):
        y1 = self.cv3(self.m(self.cv1(x)))
        y2 = self.cv2(x)
        return self.cv4(self.act(self.bn(torch.cat((y1, y2), dim=1))))


class DCNv2(torch.nn.Module):
    '''
    咩酱自实现的DCNv2，咩酱的得意之作，Pytorch的纯python接口实现，效率极高。
    '''

    def __init__(self,
                 input_dim,
                 filters,
                 filter_size,
                 stride=1,
                 padding=0,
                 bias_attr=False,
                 distribution='normal',
                 gain=1):
        super(DCNv2, self).__init__()
        assert distribution in ['uniform', 'normal']
        self.input_dim = input_dim
        self.filters = filters
        self.filter_size = filter_size
        self.stride = stride
        self.padding = padding
        self.conv_offset = torch.nn.Conv2d(input_dim, filter_size * filter_size * 3, kernel_size=filter_size,
                                           stride=stride,
                                           padding=padding, bias=True)
        # 初始化代码摘抄自SOLOv2  mmcv/cnn/weight_init.py  里的代码
        torch.nn.init.constant_(self.conv_offset.weight, 0.0)
        torch.nn.init.constant_(self.conv_offset.bias, 0.0)

        self.sigmoid = torch.nn.Sigmoid()

        self.dcn_weight = torch.nn.Parameter(torch.randn(filters, input_dim, filter_size, filter_size))
        self.dcn_bias = None
        if bias_attr:
            self.dcn_bias = torch.nn.Parameter(torch.randn(filters, ))
            torch.nn.init.constant_(self.dcn_bias, 0.0)
        if distribution == 'uniform':
            torch.nn.init.xavier_uniform_(self.dcn_weight, gain=gain)
        else:
            torch.nn.init.xavier_normal_(self.dcn_weight, gain=gain)

    def gather_nd(self, input, index):
        # 不被合并的后面的维
        keep_dims = []
        # 被合并的维
        first_dims = []
        dim_idx = []
        dims = index.shape[1]
        for i, number in enumerate(input.shape):
            if i < dims:
                dim_ = index[:, i]
                dim_idx.append(dim_)
                first_dims.append(number)
            else:
                keep_dims.append(number)

        # 为了不影响输入index的最后一维，避免函数副作用
        target_dix = torch.zeros((index.shape[0],), dtype=torch.long, device=input.device) + dim_idx[-1]
        new_shape = (-1,) + tuple(keep_dims)
        input2 = torch.reshape(input, new_shape)
        mul2 = 1
        for i in range(dims - 1, 0, -1):
            mul2 *= first_dims[i]
            target_dix += mul2 * dim_idx[i - 1]
        o = input2[target_dix]
        return o

    def forward(self, x):
        filter_size = self.filter_size
        stride = self.stride
        padding = self.padding
        dcn_weight = self.dcn_weight
        dcn_bias = self.dcn_bias

        offset_mask = self.conv_offset(x)
        offset = offset_mask[:, :filter_size ** 2 * 2, :, :]
        mask = offset_mask[:, filter_size ** 2 * 2:, :, :]
        mask = self.sigmoid(mask)

        # ===================================
        N, in_C, H, W = x.shape
        out_C, in_C, kH, kW = dcn_weight.shape
        out_W = (W + 2 * padding - (kW - 1)) // stride
        out_H = (H + 2 * padding - (kH - 1)) // stride

        # 1.先对图片x填充得到填充后的图片pad_x
        pad_x_H = H + padding * 2 + 1
        pad_x_W = W + padding * 2 + 1
        pad_x = torch.zeros((N, in_C, pad_x_H, pad_x_W), dtype=torch.float32, device=x.device)
        pad_x[:, :, padding:padding + H, padding:padding + W] = x

        # 卷积核中心点在pad_x中的位置
        rows = torch.arange(0, out_W, dtype=torch.float32, device=dcn_weight.device) * stride + padding
        cols = torch.arange(0, out_H, dtype=torch.float32, device=dcn_weight.device) * stride + padding
        rows = rows[np.newaxis, np.newaxis, :, np.newaxis, np.newaxis].repeat((1, out_H, 1, 1, 1))
        cols = cols[np.newaxis, :, np.newaxis, np.newaxis, np.newaxis].repeat((1, 1, out_W, 1, 1))
        start_pos_yx = torch.cat([cols, rows], dim=-1)  # [1, out_H, out_W, 1, 2]   仅仅是卷积核中心点在pad_x中的位置
        start_pos_yx = start_pos_yx.repeat((N, 1, 1, kH * kW, 1))  # [N, out_H, out_W, kH*kW, 2]   仅仅是卷积核中心点在pad_x中的位置
        start_pos_y = start_pos_yx[:, :, :, :, :1]  # [N, out_H, out_W, kH*kW, 1]   仅仅是卷积核中心点在pad_x中的位置
        start_pos_x = start_pos_yx[:, :, :, :, 1:]  # [N, out_H, out_W, kH*kW, 1]   仅仅是卷积核中心点在pad_x中的位置

        # 卷积核内部的偏移
        half_W = (kW - 1) // 2
        half_H = (kW - 1) // 2
        rows2 = torch.arange(0, kW, dtype=torch.float32, device=dcn_weight.device) - half_W
        cols2 = torch.arange(0, kH, dtype=torch.float32, device=dcn_weight.device) - half_H
        rows2 = rows2[np.newaxis, :, np.newaxis].repeat((kH, 1, 1))
        cols2 = cols2[:, np.newaxis, np.newaxis].repeat((1, kW, 1))
        filter_inner_offset_yx = torch.cat([cols2, rows2], dim=-1)  # [kH, kW, 2]   卷积核内部的偏移
        filter_inner_offset_yx = torch.reshape(filter_inner_offset_yx,
                                               (1, 1, 1, kH * kW, 2))  # [1, 1, 1, kH*kW, 2]   卷积核内部的偏移
        filter_inner_offset_yx = filter_inner_offset_yx.repeat(
            (N, out_H, out_W, 1, 1))  # [N, out_H, out_W, kH*kW, 2]   卷积核内部的偏移
        filter_inner_offset_y = filter_inner_offset_yx[:, :, :, :, :1]  # [N, out_H, out_W, kH*kW, 1]   卷积核内部的偏移
        filter_inner_offset_x = filter_inner_offset_yx[:, :, :, :, 1:]  # [N, out_H, out_W, kH*kW, 1]   卷积核内部的偏移

        mask = mask.permute(0, 2, 3, 1)  # [N, out_H, out_W, kH*kW*1]
        offset = offset.permute(0, 2, 3, 1)  # [N, out_H, out_W, kH*kW*2]
        offset_yx = torch.reshape(offset, (N, out_H, out_W, kH * kW, 2))  # [N, out_H, out_W, kH*kW, 2]
        offset_y = offset_yx[:, :, :, :, :1]  # [N, out_H, out_W, kH*kW, 1]
        offset_x = offset_yx[:, :, :, :, 1:]  # [N, out_H, out_W, kH*kW, 1]

        # 最终位置。其实也不是最终位置，为了更快速实现DCNv2，还要给y坐标（代表行号）加上图片的偏移来一次性抽取，避免for循环遍历每一张图片。
        start_pos_y.requires_grad = False
        start_pos_x.requires_grad = False
        filter_inner_offset_y.requires_grad = False
        filter_inner_offset_x.requires_grad = False
        pos_y = start_pos_y + filter_inner_offset_y + offset_y  # [N, out_H, out_W, kH*kW, 1]
        pos_x = start_pos_x + filter_inner_offset_x + offset_x  # [N, out_H, out_W, kH*kW, 1]
        pos_y = torch.clamp(pos_y, 0.0, H + padding * 2 - 1.0)
        pos_x = torch.clamp(pos_x, 0.0, W + padding * 2 - 1.0)
        ytxt = torch.cat([pos_y, pos_x], -1)  # [N, out_H, out_W, kH*kW, 2]

        pad_x = pad_x.permute(0, 2, 3, 1)  # [N, pad_x_H, pad_x_W, C]
        pad_x = torch.reshape(pad_x, (N * pad_x_H, pad_x_W, in_C))  # [N*pad_x_H, pad_x_W, C]

        ytxt = torch.reshape(ytxt, (N * out_H * out_W * kH * kW, 2))  # [N*out_H*out_W*kH*kW, 2]
        _yt = ytxt[:, :1]  # [N*out_H*out_W*kH*kW, 1]
        _xt = ytxt[:, 1:]  # [N*out_H*out_W*kH*kW, 1]

        # 为了避免使用for循环遍历每一张图片，还要给y坐标（代表行号）加上图片的偏移来一次性抽取出更兴趣的像素。
        row_offset = torch.arange(0, N, dtype=torch.float32, device=dcn_weight.device) * pad_x_H  # [N, ]
        row_offset = row_offset[:, np.newaxis, np.newaxis].repeat(
            (1, out_H * out_W * kH * kW, 1))  # [N, out_H*out_W*kH*kW, 1]
        row_offset = torch.reshape(row_offset, (N * out_H * out_W * kH * kW, 1))  # [N*out_H*out_W*kH*kW, 1]
        row_offset.requires_grad = False
        _yt += row_offset

        _y1 = torch.floor(_yt)
        _x1 = torch.floor(_xt)
        _y2 = _y1 + 1.0
        _x2 = _x1 + 1.0
        _y1x1 = torch.cat([_y1, _x1], -1)
        _y1x2 = torch.cat([_y1, _x2], -1)
        _y2x1 = torch.cat([_y2, _x1], -1)
        _y2x2 = torch.cat([_y2, _x2], -1)

        _y1x1_int = _y1x1.long()  # [N*out_H*out_W*kH*kW, 2]
        v1 = self.gather_nd(pad_x, _y1x1_int)  # [N*out_H*out_W*kH*kW, in_C]
        _y1x2_int = _y1x2.long()  # [N*out_H*out_W*kH*kW, 2]
        v2 = self.gather_nd(pad_x, _y1x2_int)  # [N*out_H*out_W*kH*kW, in_C]
        _y2x1_int = _y2x1.long()  # [N*out_H*out_W*kH*kW, 2]
        v3 = self.gather_nd(pad_x, _y2x1_int)  # [N*out_H*out_W*kH*kW, in_C]
        _y2x2_int = _y2x2.long()  # [N*out_H*out_W*kH*kW, 2]
        v4 = self.gather_nd(pad_x, _y2x2_int)  # [N*out_H*out_W*kH*kW, in_C]

        lh = _yt - _y1  # [N*out_H*out_W*kH*kW, 1]
        lw = _xt - _x1
        hh = 1 - lh
        hw = 1 - lw
        w1 = hh * hw
        w2 = hh * lw
        w3 = lh * hw
        w4 = lh * lw
        value = w1 * v1 + w2 * v2 + w3 * v3 + w4 * v4  # [N*out_H*out_W*kH*kW, in_C]
        mask = torch.reshape(mask, (N * out_H * out_W * kH * kW, 1))
        value = value * mask
        value = torch.reshape(value, (N, out_H, out_W, kH, kW, in_C))
        new_x = value.permute(0, 1, 2, 5, 3, 4)  # [N, out_H, out_W, in_C, kH, kW]

        # 旧的方案，使用逐元素相乘，慢！
        # new_x = torch.reshape(new_x, (N, out_H, out_W, in_C * kH * kW))  # [N, out_H, out_W, in_C * kH * kW]
        # new_x = new_x.permute(0, 3, 1, 2)  # [N, in_C*kH*kW, out_H, out_W]
        # exp_new_x = new_x.unsqueeze(1)  # 增加1维，[N,      1, in_C*kH*kW, out_H, out_W]
        # reshape_w = torch.reshape(dcn_weight, (1, out_C, in_C * kH * kW, 1, 1))  # [1, out_C,  in_C*kH*kW,     1,     1]
        # out = exp_new_x * reshape_w  # 逐元素相乘，[N, out_C,  in_C*kH*kW, out_H, out_W]
        # out = out.sum((2,))  # 第2维求和，[N, out_C, out_H, out_W]

        # 新的方案，用等价的1x1卷积代替逐元素相乘，快！
        new_x = torch.reshape(new_x, (N, out_H, out_W, in_C * kH * kW))  # [N, out_H, out_W, in_C * kH * kW]
        new_x = new_x.permute(0, 3, 1, 2)  # [N, in_C*kH*kW, out_H, out_W]
        rw = torch.reshape(dcn_weight, (
            out_C, in_C * kH * kW, 1, 1))  # [out_C, in_C, kH, kW] -> [out_C, in_C*kH*kW, 1, 1]  变成1x1卷积核
        out = F.conv2d(new_x, rw, stride=1)  # [N, out_C, out_H, out_W]
        return out


class Focus(nn.Module):
    '''
    Focus wh information into c-space from yolo4 亚像素卷积的反向操作版本
    作用: 将数据通道扩充4倍,尺寸缩小一倍,完成后进行CBS进行通道转换
    横竖每隔一个像素拿到一个值, 形成4份(2倍下采样) 再通道堆叠变成4倍通道
    每4个小方块分为一组, 每组1234号, 所有组的1展平，所有组的2展平...
    把宽度w和高度h的信息整合到c空间中
    '''

    def __init__(self, c1, c2, k=1):  # ch_in, ch_out, kernel, stride, padding, groups
        super(Focus, self).__init__()
        self.conv = FConv2d(c1 * 4, c2, k=1, act='silu')
        # self.contract = Contract(gain=2)

    def forward(self, x):  # x(b,c,w,h) -> y(b,4c,w/2,h/2)
        # 间隔为2取
        return self.conv(torch.cat([x[..., ::2, ::2], x[..., 1::2, ::2], x[..., ::2, 1::2], x[..., 1::2, 1::2]], 1))
        # return self.conv(self.contract(x))


class DeConv(nn.Module):
    ''' 反向卷积 增加尺寸 '''

    def __init__(self, in_channels, out_channels, ksize, stride=2, leaky=False):
        super(DeConv, self).__init__()
        # deconv basic config
        if ksize == 4:
            padding = 1
            output_padding = 0
        elif ksize == 3:
            padding = 1
            output_padding = 1
        elif ksize == 2:
            padding = 0
            output_padding = 0

        self.convs = nn.Sequential(
            nn.ConvTranspose2d(in_channels, out_channels, ksize, stride=stride, padding=padding,
                               output_padding=output_padding),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(0.1, inplace=True) if leaky else nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.convs(x)


def DWConv(c1, c2, k=1, s=1):
    # Depthwise convolution math.gcd() 返回的是最大公约数
    return FConv2d(c1, c2, k=k, s=s, g=math.gcd(c1, c2))
